
;
;the essential parts of the code have been implemented based on libmad 0.15 and Helix MP3 decoder
;
;do whatever you want
;


;#define atxmega

#ifdef atxmega
	.include "ATxmega128a3udef.inc"	;will also run on ATxmega256a3u
.equ xmega=1
#else
	.include "m1284pdef.inc"
;	.include "m640def.inc"			;if using ATmega 640/1280/1281/2560/2561
									;don't forget to adapt the data segment definitions at the end of this file
									;and the PROGMEM_SIZE
.equ xmega=0
#endif

#ifndef PROGMEM_SIZE
//#define PROGMEM_SIZE 131072		;use for ATmega 2560/2561
//#define PROGMEM_SIZE 65536		;use for ATmega 1280/1281
#define PROGMEM_SIZE 32768			;use for ATmega 640
#endif

;the speed<->memory usage optimization is done the easy way
;every unrolled loop remains unrolled, only the macro insertion is switched between full code or only subroutine call

;set to 0: requantization function is optimized for low flash memory consumption
;set to 1: requantization function is optimized for speed
.equ reqmode=1

;set to 0: antialiasing function is optimized for low flash memory consumption
;set to 1: antialiasing function is optimized for speed
.equ aalmode=1

;set to 0: fast sdct function is optimized for low flash memory consumption
;set to 1: fast sdct function is optimized for speed
.equ fsdctmode=1

;set to 0: sdct2 function is optimized for low flash memory consumption
;set to 1: sdct2 function is optimized for speed
.equ sdct2mode=1

;set to 0: dct4 function is optimized for low flash memory consumption
;set to 1: dct4 function is optimized for speed
.equ dctmode=1

;set to 0: imdct function is optimized for low flash memory consumption
;set to 1: imdct function is optimized for speed
.equ imdctmode=1

;set to 0: overlay function is optimized for low flash memory consumption
;set to 1: overlay function is optimized for speed
.equ ovlmode=1

;set to 0: dct32 function is optimized for low flash memory consumption
;set to 1: dct32 function is optimized for speed
.equ dct32mode=1

;set to 0: synth function is optimized for low flash memory consumption
;set to 1: synth function is optimized for speed
.equ sfmode=1

.equ round=0

;set to 1 to check how much time (in sample ticks) per mp3 frame is not consumed by the decoder
;to see just connect a terminal with bitrate "serial_br"/"serial_speed" (as defined further down) 8n1 to the serial interface
.equ speedcheck=0

;set log to 1 to "play back" mp3 from AT(x)mega flash in AVR Studio simulator
;ATmega on Simulator V1:
;the output can be written to a file using menu "Debug"->"AVR Simulator Options"->"Stimuli and logging"
;enter an output file name for PORTA
;ATxmega on Simulator V2:
;the output can be written to a file using menu "Debug"->"AVR Simulator 2 Options",
;checking "Run Stimuli File" and selecting the file "Portalog.stim"; this will write output into "output.log" file
;delete the "output.log" file before running code again to avoid inconsistent data
;note that Simulator V2 is running the code WAY (really) slower that V1
.equ log=1

;if playing back mp3 using PWM the output will be only 8 bit samples and 12 bit if DAC output is used on ATxmega
;set chip to 1 to play back mp3 from ATmega flash
.equ chip=0

;set flash to 1 to play back mp3 from an external SPI flash
.equ flash=0
;if using this option the mp3 file must be preceeded by a two-byte value (low byte first) holding the number of mp3 frames
;to calculate use a tool like mp3guessenc

;to avoid the need of a double buffered output the polyphase filter calculation will be suspended until the play back routine
;reaches the nxtfrmtrg offset; only used if no support of mp3 header evaluation is chosen
;the value may have to be adapted depending on clock frequency and speed/memory optimization
	.if chip==1 || flash==1
.equ nxtfrmtrg=512
	.endif

#define clock 16;MHz			CPU clock frequency

;basic setting in case no support of mp3 header evaluation is chosen
.equ bitrate=16;kbit/s			bit rate

;sampling frequency for PWM output, also used with DAC output if no support of mp3 header evaluation is chosen
.equ sfrq=8;kHz                sampling frequency; use 8,11 or 12 kHz

;set to 1 to use built-in DAC on ATxmega
.equ use_dac=0

;set multiplicator of sampling frequency (only used if use_dac set to 1)
;can be used to play back audio with double or quadruple speed
.equ sfrq_mul=1					;1, 2 or 4 are supported

;set to 1 to support mp3 header evaluation and short blocks including double buffered output as well
;if set to 0 only long blocks are supported, use Lame with flag --noshort when converting audio files to mp3
.equ eval_hd=1

;set to 1 to support the bit reservoir
;if set to 0 use Lame with flag --nores when converting audio files to mp3
.equ br_support=1
.equ fr_buf_size=1024			;size of mp3 frame buffer needed for support of bit reservoir, set to multiples of 1024

	.if eval_hd==1
	.if xmega==1
.equ btype=GPIO_GPIOR0
	.else
.equ btype=GPIOR0
	.endif 
.equ altwnd=0					;flag for alternate window
.equ bt2=1						;flag for block type 2 (short block)
.equ bt3=2						;flag for block type 3 (stop block)
.equ latch_clk=6				;flag for locked sample clock (only for use with DAC)
.equ cbuf=7						;flag for current sample buffer
	.endif

;define GPIO port for zero value (used in PWM interrupt)
	.if xmega==1
.equ pzero=GPIO_GPIOR1
	.else
.equ pzero=GPIOR1
	.endif

;define /SS pin for SPI interface to serial flash
#ifdef atxmega
#define SPI_SS PORTC,4
#else
#define SPI_SS PORTB,4
#endif

;audio PWM should run at full clock speed
;the oversampling counter should be calculated so that the resulting output frequency is close to the
;mp3 sampling frequency as much as possible
.equ ovrsmpl=int(((clock*10000)/(sfrq*256)+5)/10)

.equ frame_size=72*bitrate/sfrq		;mp3 frame size, only used if no support of mp3 header evaluation is chosen

	.if chip ==1 || flash==1
	.if xmega==1
.equ serial_br=115200;bit/s			;for ATxmega

.set bsel=int((clock*62500*128)/serial_br)-128
.set bscale=-7
	.if bsel>4095
.set bsel=int((clock*62500*64)/serial_br)-64
.set bscale=-6
	.endif
	.if bsel>4095
.set bsel=int((clock*62500*32)/serial_br)-32
.set bscale=-5
	.endif
	.if bsel>4095
.set bsel=int((clock*62500*16)/serial_br)-16
.set bscale=-4
	.endif
	.else
.equ serial_speed=16				;for ATmega, 115k2@16 MHz clock, X2 clockmode
	.endif

.equ	flash_page_pgm=2
.equ	flash_rd_data=3
.equ	flash_rd_sr1=5
.equ	flash_wr_enable=6
.equ	flash_sect_erase=$20
	.endif

#ifdef atxmega
;Pin definitions for serial interface
#define SERIALPORT		PORTC
#define SERIALPORT_dd	DDRC
#define SERIALPORT_TX	3
.equ    SERIAL_IF	   =USARTC0_base
.equ    SERIAL_IF_INTV =USARTC0_vbase
#endif

#ifdef atxmega
.equ PORTA=VPORT0_OUT
.equ PINA=VPORT0_IN
.equ DDRA=VPORT0_DIR
.equ PORTB=VPORT1_OUT
.equ PINB=VPORT1_IN
.equ DDRB=VPORT1_DIR
.equ PORTC=VPORT2_OUT
.equ PINC=VPORT2_IN
.equ DDRC=VPORT2_DIR
.equ PORTD=VPORT3_OUT
.equ PIND=VPORT3_IN
.equ DDRD=VPORT3_DIR
#endif

	.if sfrq_mul!=1 && sfrq_mul!=2 && sfrq_mul!=4
	.error "***********************************"
	.error "sfrq_mul out of range; use 1,2 or 4"
	.error "***********************************"
	.exit
	.endif

	.if SRAM_SIZE<8192
	.error "*****************************************"
	.error "Choose device with at least 8 kBytes SRAM"
	.error "*****************************************"
	.exit
	.endif

	.ifndef DACB_CTRLA
	.if use_dac==1
	.error "****************************************"
	.error "Selected device doesn't have builtin DAC"
	.error "****************************************"
	.exit
	.endif
	.endif

	.if eval_hd==0 && sfrq==11
	.error "*****************************************************************************"
	.error "Only 8 or 12 kHz sample frequency is allowed if header evaluation is disabled"
	.error "*****************************************************************************"
	.exit
	.endif

	.macro rd_byte 
	.if flash==1
rd_byte_wt:
	.if xmega==1
	lds r1,(SPIC_STATUS)
	sbrs r1,SPI_IF_bp
	rjmp rd_byte_wt			;wait if byte has not been read completely
	lds r1,(SPIC_DATA)
	sts (SPIC_DATA),r1		;trigger the read of next byte
	.else
	in r1,SPSR
	sbrs r1,SPIF
	rjmp rd_byte_wt			;wait if byte has not been read completely
	in r1,SPDR
	out SPDR,r1				;trigger the read of next byte
	.endif
	.if br_support==0
	adiw Z,1
	.endif
	.else
	.if FLASHEND>$7fff
	elpm r1,Z+
	.else
	lpm r1,Z+
	.endif
	.endif
	.endmacro

	.macro rd_buf
	ld r1,Z+
;in order to make sure the circular buffer handling will work correctly, the frame_buf must start on an address
;where the first fixed bit is zero
;example for fr_buf_size=1024:
;frame_buf=$1000 would be OK, it's   0001.0>0<XX.XXXX.XXXX
;frame_buf=$1400 will not work, it's 0001.0>1<XX.XXXX.XXXX
;any additional effort for further address handling might impact the decoder's performance significantly
	andi ZH,HIGH(frame_buf+fr_buf_size-1)
	.endmacro

	.macro rd_bbyte
	.if br_support==1
	rd_buf
	.else
	rd_byte
	.endif
	.endmacro

	.cseg
	.org 0
	.if chip==1 || flash==1
	rjmp init
	.endif

#ifdef atxmega
	.if flash==1
	.org SERIAL_IF_INTV+USART_RXC_voffset
	rjmp rcv_int
	.endif

	.if chip==1 || flash==1
	.org TCD0_OVF_vect
	rjmp aint
	.endif
#else
	.if chip==1 || flash==1
	.org OVF1addr
	rjmp aint
	.endif

	.if flash==1
	.org URXC0addr
	rjmp rcv_int
	.endif
#endif

	.if chip==1 || flash==1
init:
	.endif
	.if xmega==1
	ldi r17,CCP_IOREG_gc
	out CPU_CCP,r17
	ldi r16,OSC_RC32MEN_bm						;enable 32Mhz internal clock
	sts (OSC_CTRL),r16
wait_for_osc: lds r16,(OSC_STATUS)
	andi r16,OSC_RC32MRDY_bm
	breq wait_for_osc

	ldi r16,1
	sts (DFLLRC32M_CTRL),r16
	ldi r16,OSC_PLLSRC_RC32M_gc+int(clock/8)	;Source internal 32MHz/4
	sts (OSC_PLLCTRL),r16

	out CPU_CCP,r17
	ldi r16,OSC_PLLEN_bm+OSC_RC32MEN_bm			;enable PLL + 32MHz internal clock
	sts (OSC_CTRL),r16
wait_for_osc1: lds r16,(OSC_STATUS)
	andi r16,OSC_PLLRDY_bm
	breq wait_for_osc1	

	out CPU_CCP,r17
	ldi r16,CLK_SCLKSEL_PLL_gc					;select PLL as clock source
	sts (CLK_CTRL),r16

;assign PORT A, B, C and D to virtual ports in ATMega manner
	ldi r16,PORTCFG_VP13MAP_PORTB_gc|PORTCFG_VP02MAP_PORTA_gc
	sts (PORTCFG_VPCTRLA),r16
	ldi r16,PORTCFG_VP13MAP_PORTD_gc|PORTCFG_VP02MAP_PORTC_gc
	sts (PORTCFG_VPCTRLB),r16
	.else
	.if br_support==1
	ldi r16,LOW(frame_buf-1)
	ldi r17,HIGH(frame_buf-1)
	.else
	ldi r16,LOW(RAMEND)
	ldi r17,HIGH(RAMEND)
	.endif
	out SPL,r16
	out SPH,r17
	.endif

	.if log==1
	ldi r16,-1
	out DDRA,r16
	out PORTA,r16
	com r16
	out PORTA,r16
	com r16
	out PORTA,r16
	.endif

	ldi XL,LOW(ovlb)
	ldi XH,HIGH(ovlb)
	ldi YL,LOW(lo)
	ldi YH,HIGH(lo)
	ldi ZL,LOW(hi)
	ldi ZH,HIGH(hi)
	clr r15					;r15 needs to remain 0 everywhere
	ldi r17,576/4
clr_buffers: st X+,r15		;initialize overlay buffer and polyphase filter buffer
	st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Z+,r15
	st Z+,r15
	st Z+,r15
	st Z+,r15
	dec r17
	brne clr_buffers
	sts (phase),r15

	.if br_support==1
	sts (fr_wr_ptr),r15
	sts (fr_wr_ptr+1),r15
	.endif

	.if flash==1
;set up SPI interface towards serial flash
	.if xmega==1
	ldi r16,0b10111000				;CLK, MOSI, /SS and TxD for output
	out PORTC,r16
	out DDRC,r16
	ldi r16,SPI_CLK2X_bm|SPI_ENABLE_bm|SPI_MASTER_bm|SPI_MODE_0_gc|SPI_PRESCALER_DIV4_gc	;SPI enabled as master
	sts (SPIC_CTRL),r16
	lds r16,(PORTE_IN)
	sbrs r16,5						;holding PINE.5 on L will jump to the serial flash loader
	jmp load_flash
	.else
	sbi SPI_SS						;/SS for serial flash
	ldi r16,1<<DDB7|1<<DDB5|1<<DDB4	;set port mode for SCLK, MOSI and /SS to output
	out DDRB,r16
	ldi r16,1<<SPE|1<<MSTR			;SPI enabled as master
	out SPCR,r16
	ldi r16,1<<SPI2X				;SPI clock frequency set to fclk/2
	out SPSR,r16
	sbic PIND,3						;holding PIND.3 on H will jump to the serial flash loader
	jmp load_flash
	.endif

;set up serial interface
	.if xmega==1 && speedcheck==1
	ldi YL,LOW(SERIAL_IF)
	ldi YH,HIGH(SERIAL_IF)
	ldi r16,LOW(bsel)
	std Y+USART_BAUDCTRLA_offset,r16
	ldi r16,bscale<<4|HIGH(bsel)
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_CMODE_ASYNCHRONOUS_gc|USART_CHSIZE_8BIT_gc	;async 8n1
	std Y+USART_CTRLC_offset,r16
	ldi r16,USART_RXEN_bm|USART_TXEN_bm
	std Y+USART_CTRLB_offset,r16
	ldi r16,USART_RXCIF_bm
	std Y+USART_STATUS_offset,r16
	.endif
	.if xmega!=1 && speedcheck==1
	clr r16
	sts UBRR0H,r16
	ldi r16,serial_speed
	sts UBRR0L,r16
	ldi r16,1<<TXC0|1<<U2X0
	sts UCSR0A,r16
	ldi r16,1<<RXEN0|1<<TXEN0
	sts UCSR0B,r16
	ldi r16,1<<USBS0|1<<UCSZ01|1<<UCSZ00	;8n1
	sts UCSR0C,r16
	.endif
	.endif

	.if chip==1 || flash==1
	out pzero,r15				;set GPIO port to 0, needed in PWM interrupt
	.endif

	.if chip==1 || flash==1
	.if xmega==1
	ldi r16,PMIC_HILVLEN_bm|PMIC_MEDLVLEN_bm|PMIC_LOLVLEN_bm	;enable all interrupt levels
	sts (PMIC_CTRL),r16
	.endif

;set up timer for audio playback (either for PWM or DAC output)
	.if use_dac==1
	ldi r16,TC_CLKSEL_DIV1_gc
	sts (TCD0_CTRLA),r16
	ldi r16,TC_WGMODE_NORMAL_gc
	sts (TCD0_CTRLB),r16
	ldi r16,LOW(clock*(1000/sfrq_mul)/sfrq)
	sts (TCD0_PER),r16
	ldi r16,HIGH(clock*(1000/sfrq_mul)/sfrq)
	sts (TCD0_PER+1),r16
	ldi r16,TC_OVFINTLVL_LO_gc
	sts (TCD0_INTCTRLA),r16

	ldi r16,DAC_CH1EN_bm|DAC_CH0EN_bm|DAC_ENABLE_bm
	sts (DACB_CTRLA),r16
	ldi r16,DAC_CHSEL_DUAL_gc
	sts (DACB_CTRLB),r16
	ldi r16,DAC_LEFTADJ_bm
	sts (DACB_CTRLC),r16
	.else
	.if xmega==1
	ldi r16,3
	out DDRD,r16
	ldi r16,TC_CLKSEL_DIV1_gc
	sts (TCD0_CTRLA),r16
	ldi r16,TC0_CCAEN_bm|TC0_CCBEN_bm|TC_WGMODE_SINGLESLOPE_gc
	sts (TCD0_CTRLB),r16
	ldi r16,255
	sts (TCD0_PER),r16
	sts (TCD0_PER+1),r15
	ldi r16,TC_OVFINTLVL_LO_gc
	sts (TCD0_INTCTRLA),r16

	ldi r16,$80
	sts (TCD0_CCA),r16
	sts (TCD0_CCA+1),r15
	sts (TCD0_CCB),r16
	sts (TCD0_CCB+1),r15
	.else
	ldi r16,1<<DDD5|1<<DDD4		;set port mode for OC1A/OC1B to output
	out DDRD,r16
	ldi r16,$80
	sts OCR1AL,r16
	sts OCR1BL,r16
	ldi r16,1<<COM1A1|1<<COM1B1|1<<COM1B0|1<<WGM10	;OC1A direct, OC1B inverted output, Fast PWM 8 bit
	sts TCCR1A,r16
	ldi r16,1<<WGM12|1<<CS10	;Fast PWM 8 bit, no prescaler
	sts TCCR1B,r16
	ldi R16,1<<TOIE1			;enable interrupt at timer overflow
	sts TIMSK1,r16
	.endif

	ldi r16,1
	mov r14,r16					;r14 = oversampling counter
	.endif

	.if xmega==1
	ldi r16,SLEEP_SMODE_IDLE_gc|SLEEP_SEN_bm
	sts (SLEEP_CTRL),r16		;prepare idle mode activation
	.else
	ldi r16,1<<SE
	out SMCR,r16				;prepare idle mode activation
	.endif

;initialize output buffer to audio "0"
	ldi XL,LOW(output)
	ldi XH,HIGH(output)
	sts (outptr),XL
	sts (outptr+1),XH
	ldi r16,576/4
	ldi r17,$80
clr_output:
	.if use_dac==1
	st X+,r15
	st X+,r17
	st X+,r15
	st X+,r17
	st X+,r15
	st X+,r17
	st X+,r15
	st X+,r17
	.else
	st X+,r17
	st X+,r17
	st X+,r17
	st X+,r17
	.endif
	dec r16
	brne clr_output
	.if eval_hd==1
	out btype,r15
	.endif
	sei
	.endif

	.if flash==1
	clr ZL
	clr ZH
	clr r23
	cbi SPI_SS
	ldi r16,flash_rd_data
	rcall wr_spi
	push ZL
	push ZH
	push r23
	rcall set_addr
	rcall rd_spi
	mov r24,r16
	rcall rd_spi
	mov r25,r16			;r25:r24 - mp3 frame count
	sbi SPI_SS
	ldi r16,-1
	cp r24,r16
	cpc r25,r16
	brne playback
	rjmp end			;if frame count==$ffff assume empty flash
playback: ldi ZL,2			;mp3 data starts at flash address 000002
	.else
	ldi ZL,LOW(mp3file*2)
	ldi ZH,HIGH(mp3file*2)
	.ifdef RAMPZ
	ldi r23,PAGE(mp3file*2)
	out RAMPZ,r23
	.endif
	.ifdef CPU_RAMPZ
	ldi r23,PAGE(mp3file*2)
	out CPU_RAMPZ,r23
	.endif
	ldi r24,LOW((mp3file_end-mp3file)*2/frame_size)
	ldi r25,HIGH((mp3file_end-mp3file)*2/frame_size)
	.endif

eval_frame:
	.if flash==1
	push r24
	push r25
	.endif
	push ZL
	push ZH
	.if log==1 || chip==1
	.ifdef RAMPZ
	push r23
	.endif
	.ifdef CPU_RAMPZ
	push r23
	.endif
	.endif

	.if flash==1
	push r23
	mov r16,r23
	cbi SPI_SS
	ldi r16,flash_rd_data
	rcall wr_spi
	mov r16,r23
	rcall wr_spi
	mov r16,ZH
	rcall wr_spi
	mov r16,ZL
	rcall wr_spi
	.if xmega==1
	sts (SPIC_DATA),r1			;trigger the reading of next byte
	.else
	out SPDR,r1					;trigger the reading of next byte
	.endif
	rcall rd_byte1
	mov r20,r1
	rcall rd_byte1
	mov r21,r1
	rcall rd_byte1
	mov r22,r1
	rcall rd_byte1
	mov r23,r1
	.else
	.if FLASHEND>$7fff
	elpm r20,Z+
	elpm r21,Z+
	elpm r22,Z+
	elpm r23,Z+
	.else
	lpm r20,Z+
	lpm r21,Z+
	lpm r22,Z+
	lpm r23,Z+
	.endif
	.endif
	.if eval_hd
	cpi r20,'I'
	brne no_tag
	cpi r21,'D'
	brne no_tag
	cpi r22,'3'
	brne no_tag
;one may handle ID3 tag here
;only tags up from v2 might be inserted in front of the mp3 data
	rcall rd_byte1				;skip revision number
	rcall rd_byte1				;skip flags
	rcall rd_byte1				;skip length byte
	rcall rd_byte1				;skip length byte
	rcall rd_byte1
	mov r17,r1
	clr r16
	lsr r17
	ror r16
	rcall rd_byte1
	add r16,r1					;only tags with a maximum length of 16383 bytes are handled
	subi r16,-10				;add 10 for tag header
	sbci r17,-1
	sts (frame_sz),r16
	sts (frame_sz+1),r17
	rjmp skip_frame
no_tag:
	ldi r16,0b00011111			;mask for MPEG version ID, Layer descr. and Protection bit
	and r16,r21
	cpi r16,0b00000011			;only MPEG Version 2.5 Layer III and no protection supported yet
	brne skip_frm_jmp1
	ldi r16,0b11000011			;mask for Channel Mode and Emphasis
	and r16,r23
	cpi r16,0b11000000			;only Single channel and no emphasis supported yet
	brne skip_frm_jmp1
	ldi r16,0b11110000			;mask for Bitrate index
	and r16,r22
	breq skip_frm_jmp1			;no support for free bitrate
	swap r16
	cpi r16,9					;max. 64 kbit/s supported
	brcc skip_frm_jmp1
	dec r16
	add r16,r16
	ldi r19,0b00001100			;mask for Sampling frequency index
	and r19,r22
	lsr r19
	cpi r19,6					;check for "reserved"
	brne sfrq_ok
skip_frm_jmp1: rjmp skip_frame
sfrq_ok: movw r7:r6,Z
	.if use_dac==1
	sbic btype,latch_clk
	rjmp skip_ssfrq
	sbi btype,latch_clk
	ldi ZL,LOW(sfrq_tbl*2)
	ldi ZH,HIGH(sfrq_tbl*2)
	add ZL,r19
	adc ZH,r15
	lpm r2,Z+
	lpm r4,Z
	sts (TCD0_PER),r2
	sts (TCD0_PER+1),r4
skip_ssfrq:
	.endif
	ldi r18,8					;size of table part for each sampling frequency
	mul r18,r19
	add r16,r0
	ldi ZL,LOW(frm_sz_tbl*2)
	ldi ZH,HIGH(frm_sz_tbl*2)
	add ZL,r16
	adc ZH,r15
	lpm r16,Z+
	sbrc r22,1					;skip if padding bit not set
	inc r16						;no carry bit handling as no carry to upper byte is expected
	sts (frame_sz),r16
	lpm r16,Z
	sts (frame_sz+1),r16
	movw Z,r7:r6
	.endif

	ldi YL,LOW(part23len)
	ldi YH,HIGH(part23len)
	rd_byte
	ldi r18,8
	ldi r16,8
	rcall get_bits		;Main_data_begin
	.if br_support==1
	lds r16,(fr_wr_ptr)
	lds r17,(fr_wr_ptr+1)
	sub r16,r2
	sbc r17,r15
	brcc set_fr_rd_ptr
	subi r16,LOW(-fr_buf_size)
	sbci r17,HIGH(-fr_buf_size)
set_fr_rd_ptr: sts (fr_rd_ptr),r16
	sts (fr_rd_ptr+1),r17
	.else
	tst r2
	brne skip_frm_jmp	;skip frame if referring backwards
	.endif
	ldi r16,1
	rcall get_bits		;skip Private_bits
	ldi r16,12
	rcall get_bits		;Part2_3_length
	st Y+,r2
	st Y+,r3
	or r2,r3
	brne rd_head1
skip_frm_jmp: rjmp skip_frame		;skip if zero (most likely last frame)
rd_head1: ldi r16,9
	rcall get_bits		;Big_values
	st Y+,r2
	st Y+,r3
	ldi r16,8
	rcall get_bits		;Global_gain
	st Y+,r2
	ldi r16,9
	rcall get_bits		;Scalefac_compress
	st Y+,r2
	st Y+,r3
	ldi r16,1
	rcall get_bits		;Window_switching
	tst r2
	.if eval_hd!=1
	brne skip_frm_jmp	;skip short/mixed block
	.else
	brne alt_blk
	cbi btype,altwnd
	cbi btype,bt2
	cbi btype,bt3
	.endif
	ldi r16,5
	rcall get_bits		;Table[0] select
	st Y+,r2
	ldi r16,5
	rcall get_bits		;Table[1] select
	st Y+,r2
	ldi r16,5
	rcall get_bits		;Table[2] select
	st Y+,r2
	ldi r16,4
	rcall get_bits		;Region0_count
	st Y+,r2
	ldi r16,3
	rcall get_bits		;Region1_count
	st Y+,r2
	.if eval_hd==1
	movw r5:r4,Z
	rjmp w_common
alt_blk: 
	sbi btype,altwnd
	cbi btype,bt2
	cbi btype,bt3
	ldi r16,3
	rcall get_bits		;Block_type (Bit 1...2) + Mixed_block (Bit 0)
	sbrc r2,0
	rjmp skip_frame		;skip frame if mixed block (very (!) unlikely)
	mov r17,r2
	cpi r17,2<<1		;check if Block_type==2
	brne alt_blk1
	sbi btype,bt2		;set bit for Block_type 2
alt_blk1: cpi r17,3<<1
	brne alt_blk2
	sbi btype,bt3		;set bit for Block_type 3
alt_blk2:
	ldi r16,7
	sbic btype,bt2		;skip if not Block_type 2
	inc r16
	sts (r0cnt),r16
	ldi r17,37
	sbis btype,bt2		;skip if Block_type 2
	ldi r17,20
	sub r17,r16
	sts (r1cnt),r17
	ldi r16,5
	rcall get_bits		;Table [0] select
	st Y+,r2
	ldi r16,5
	rcall get_bits		;Table [1] select
	st Y+,r2
						;Table [2] select always 0
	ldi YL,LOW(sblk_gain)
	ldi YH,HIGH(sblk_gain)
	ldi r16,3
	rcall get_bits		;Sub_block_gain [0]
	st Y+,r2
	ldi r16,3
	rcall get_bits		;Sub_block_gain [1]
	st Y+,r2
	ldi r16,3
	rcall get_bits		;Sub_block_gain [2]
	st Y,r2
	ldi YL,LOW(scfacsc)
	ldi YH,HIGH(scfacsc)
	movw r5:r4,Z
	ldi ZL,LOW(sfbs_tbl*2)
	ldi ZH,HIGH(sfbs_tbl*2)
	sbic btype,bt2		;skip if not Block_type 2
	rjmp w_common1
w_common:
	ldi ZL,LOW(sfbl_tbl*2)
	ldi ZH,HIGH(sfbl_tbl*2)
w_common1: add ZL,r19
	adc ZH,r15
	lpm r2,Z+
	lpm r3,Z
	sts (sfb_ptr),r2
	sts (sfb_ptr+1),r3
	movw Z,r5:r4
	.endif
	ldi r16,1
	rcall get_bits		;Scalefac_scale
	st Y+,r2
	ldi r16,1
	rcall get_bits		;Count1_table_select
	st Y,r2
	.if br_support==0
	mov r0,r1			;copy remaining bits
	.endif

	.if br_support==1
	.if eval_hd==1
	lds r24,(frame_sz)
	lds r25,(frame_sz+1)
	.else
	ldi r24,LOW(frame_size)
	ldi r25,HIGH(frame_size)
	.endif
	subi r24,4+9		;size of header (32 bit) + size of side info (72 bit)
	sbc r25,r15
	lds YL,(fr_wr_ptr)
	lds YH,(fr_wr_ptr+1)
	subi YL,LOW(-frame_buf)
	sbci YH,HIGH(-frame_buf)
	ldi r17,HIGH(frame_buf+fr_buf_size)
	rjmp read_frame1
read_frame: rd_byte
read_frame1: st Y+,r1
	cpi YL,LOW(frame_buf+fr_buf_size)
	cpc YH,r17
	brcs read_frame2
	ldi YL,LOW(frame_buf)
	ldi YH,HIGH(frame_buf)
read_frame2: sbiw r25:r24,1
	brne read_frame
	subi YL,LOW(frame_buf)
	sbci YH,HIGH(frame_buf)
	sts (fr_wr_ptr),YL
	sts (fr_wr_ptr+1),YH
	lds ZL,(fr_rd_ptr)
	lds ZH,(fr_rd_ptr+1)
	subi ZL,LOW(-frame_buf)
	sbci ZH,HIGH(-frame_buf)
	rd_bbyte
	mov r0,r1
	.if flash==1
	sbi SPI_SS
	.endif
	.endif
	rcall decode_frame

	.if flash==1 && br_support==0
	sbi SPI_SS
	.endif

	rcall requantize

	.if eval_hd==1
	sbic btype,bt2		;skip if not Block_type 2
	rcall reorder
	sbis btype,bt2		;skip if Block_type 2
	.endif
	rcall antialiase
	call transform

	.if (chip==1 || flash==1) && eval_hd!=1
	.if use_dac==1
	ldi r18,HIGH(output+2*nxtfrmtrg)
	.else
	ldi r18,HIGH(output+nxtfrmtrg)
	.endif
	.if (chip==1 || flash==1) && speedcheck==1
	sleep						;wait for next interrrupt to synchronize
	lds XL,(outptr)
	lds XH,(output+1)
	.if use_dac==1
	ldi r16,LOW(output+2*nxtfrmtrg)
	.else
	ldi r16,LOW(output+nxtfrmtrg)
	.endif
	sub r16,XL
	mov XL,r16
	.if use_dac==1
	ldi r16,HIGH(output+2*nxtfrmtrg)
	.else
	ldi r16,HIGH(output+nxtfrmtrg)
	.endif
	sbc r16,XH
	.if use_dac==1
	lsr r16
	ror XL
	.endif
	push r16
	push XL
	.endif
sync_lp: sleep					;save some milliwatts while waiting for next interrupt
	lds r16,(outptr)
	lds r17,(outptr+1)
	.if use_dac==1
	cpi r16,LOW(output+2*nxtfrmtrg)
	.else
	cpi r16,LOW(output+nxtfrmtrg)
	.endif
	cpc r17,r18
	brcs sync_lp
	.endif

	call synthf
	.if log==1
	rcall sample_dump
	.endif

	.if (chip==1 || flash==1) && speedcheck==1 && eval_hd!=1
	pop r16
	rcall disp_hex
	pop r16
	rcall disp_hex
	ldi r16,13
	rcall disp_nibble1
	ldi r16,10
	rcall disp_nibble1
	.endif

skip_frame: 

	.if flash==1
	rcall spi_off
	pop r23
	.endif
	.if log==1 || chip==1
	.ifdef RAMPZ
	pop r23
	.endif
	.ifdef CPU_RAMPZ
	pop r23
	.endif
	.endif

	pop ZH
	pop ZL
	.if eval_hd==1
	lds r16,(frame_sz)
	lds r17,(frame_sz+1)
	add ZL,r16
	adc ZH,r17
	.if flash==1
	adc r23,r15
	.endif
	.if log==1 || chip==1
	.ifdef RAMPZ
	adc r23,r15
	out RAMPZ,r23
	.endif
	.ifdef CPU_RAMPZ
	adc r23,r15
	out CPU_RAMPZ,r23
	.endif
	.endif
	.else
	subi ZL,LOW(-frame_size)
	sbci ZH,HIGH(-frame_size)
	.if flash==1
	sbci r23,-1
	.endif
	.if log==1 || chip==1
	.ifdef RAMPZ
	sbci r23,-1
	out RAMPZ,r23
	.endif
	.ifdef CPU_RAMPZ
	sbci r23,-1
	out CPU_RAMPZ,r23
	.endif
	.endif
	.endif

	.if flash==1
	pop r25
	pop r24
	.endif

	.if flash==1
	sbiw r25:r24,1
	breq end
	.else
	cpi ZL,LOW(mp3file_end*2)
	ldi r16,HIGH(mp3file_end*2)
	cpc ZH,r16
	.ifdef RAMPZ
	ldi r16,PAGE(mp3file_end*2)
	cpc r23,r16
	.endif
	.ifdef CPU_RAMPZ
	ldi r16,PAGE(mp3file_end*2)
	cpc r23,r16
	.endif
	brcc end
	.endif

	.if (chip==1 || flash==1) && eval_hd==1
	in r17,btype
	andi r17,1<<cbuf
	.if speedcheck==1
	sleep							;wait for next interrupt to synchronize
	lds XL,(outptr)
	lds XH,(outptr+1)
	.if use_dac==1
	ldi r16,LOW(output+2*575)
	sbic btype,cbuf
	ldi r16,LOW(output1+2*575)
	sub r16,XL
	mov XL,r16
	ldi r16,HIGH(output+2*575)
	sbic btype,cbuf
	ldi r16,HIGH(output1+2*575)
	sbc r16,XH
	mov XH,r16
	lsr XH
	ror XL
	.else
	ldi r16,LOW(output+575)
	sbic btype,cbuf
	ldi r16,LOW(output1+575)
	sub r16,XL
	mov XL,r16
	ldi r16,HIGH(output+575)
	sbic btype,cbuf
	ldi r16,HIGH(output1+575)
	sbc r16,XH
	.endif
	rcall disp_hex
	mov r16,XL
	rcall disp_hex
	ldi r16,13
	rcall disp_nibble1
	ldi r16,10
	rcall disp_nibble1
	.endif
sync_lp: sleep						;save some milliwatts while waiting for next interrupt
	in r16,btype
	andi r16,1<<cbuf
	cp r16,r17
	breq sync_lp
	.endif
	rjmp eval_frame
end: 
	.if chip==1 || flash==1
	cli
	.if xmega==1
	ldi r16,SLEEP_SMODE_PDOWN_gc|SLEEP_SEN_bm	;enable power down
end_lp: sts (SLEEP_CTRL),r16
	sleep
	rjmp end_lp							;just in case...
	.else
	ldi r16,1<<SM1|1<<SE				;enable power down
end_lp: out SMCR,r16
	sleep
	rjmp end_lp							;just in case...
	.endif
	.else
end_lp: rjmp end_lp						;set a breakpoint here in simulator
	.endif

rd_byte1: rd_byte
	ret

	.if flash==1
spi_off:
	.if xmega==1
	lds r1,(SPIC_STATUS)	;wait if SPI transfer is still in progress
	sbrs r1,SPI_IF_bp
	rjmp spi_off
	lds r1,(SPIC_DATA)
	.else
	in r1,SPSR				;wait if SPI transfer is still in progress
	sbrs r1,SPIF
	rjmp spi_off
	in r1,SPDR
	.endif
	sbi SPI_SS				;deselect SPI flash
	ret
	.endif

	.if log==1
sample_dump:
	.if eval_hd==1
	ldi ZL,LOW(output1)
	ldi ZH,HIGH(output1)
	.else
	ldi ZL,LOW(output)
	ldi ZH,HIGH(output)
	.endif
	ldi r24,LOW(576)
	ldi r25,HIGH(576)
sample_dump_lp: ld r16,Z+
	push r16
	ld r16,Z+
	push r16
	call dumpbyte
	call dumpbyte
	ldi r16,3
	out PORTA,r16
	sbiw r25:r24,1
	brne sample_dump_lp
	ret
	.endif

get_bits:
	clr r2
	clr r3
get_bits_lp: lsl r1
	rol r2
	rol r3
	dec r18
	breq get_bits_nxt
get_bits1: dec r16
	brne get_bits_lp
	ret
get_bits_nxt: rd_byte
	ldi r18,8
	rjmp get_bits1

;audio timer interrupt
	.if chip==1 || flash==1
	.if use_dac==1
aint: in r15,CPU_SREG
	push ZL
	push ZH
	push r16
	push r14
	lds ZL,(outptr)
	lds ZH,(outptr+1)
	ld r14,Z+
	ld r16,Z+
	sts (DACB_CH0DATA),r14
	sts (DACB_CH0DATA+1),r16
	sts (DACB_CH1DATA),r14
	sts (DACB_CH1DATA+1),r16
	ldi r16,HIGH(output+2*576)
	.if eval_hd==1
	sbic btype,cbuf
	ldi r16,HIGH(output1+2*576)
	.endif
	cpi ZL,LOW(output+2*576)
	.if eval_hd==1
	sbic btype,cbuf
	cpi ZL,LOW(output1+2*576)
	.endif
	cpc ZH,r16
	brne aint1
	ldi ZL,LOW(output)
	ldi ZH,HIGH(output)
	.if eval_hd==1
	sbis btype,cbuf
	ldi ZL,LOW(output1)
	sbis btype,cbuf
	ldi ZH,HIGH(output1)
	ldi r16,$80
	in r14,btype
	eor r14,r16
	out btype,r14
	.endif
aint1: sts (outptr),ZL
	sts (outptr+1),ZH
	pop r14
	pop r16
	pop ZH
	pop ZL
	out CPU_SREG,r15
	in r15,pzero
	reti
	.else
;for PWM
aint:
	.if xmega==1
	in r15,CPU_SREG
	.else
	in r15,SREG
	.endif
	dec r14
	breq aint1
	.if xmega==1
	out CPU_SREG,r15
	.else
	out SREG,r15
	.endif
	in r15,pzero		;r15 has to be always set to 0
	reti
aint1: push ZL
	push ZH
	push r16
	lds ZL,(outptr)
	lds ZH,(outptr+1)
	ld r14,Z+
	.if xmega==1
	clr r16
	sts (TCD0_CCA),r14
	sts (TCD0_CCA+1),r16
	sts (TCD0_CCB),r14
	sts (TCD0_CCB+1),r16
	.else
	sts OCR1AL,r14
	sts OCR1BL,r14
	.endif
	.if eval_hd==1
	sbic btype,cbuf
	rjmp aint2
	.endif
	ldi r16,HIGH(output+576)
	cpi ZL,LOW(output+576)
	cpc ZH,r16
	brne aint3
	.if eval_hd!=1
	ldi ZL,LOW(output)
	ldi ZH,HIGH(output)
	.else
	ldi ZL,LOW(output1)
	ldi ZH,HIGH(output1)
	sbi btype,cbuf
	rjmp aint3
aint2: ldi r16,HIGH(output1+576)
	cpi ZL,LOW(output1+576)
	cpc ZH,r16
	brne aint3
	ldi ZL,LOW(output)
	ldi ZH,HIGH(output)
	cbi btype,cbuf
	.endif
aint3: sts (outptr),ZL
	sts (outptr+1),ZH
	ldi r16,ovrsmpl
	mov r14,r16
	pop r16
	pop ZH
	pop ZL
	.if xmega==1
	out CPU_SREG,r15
	.else
	out SREG,r15
	.endif
	in r15,pzero
	reti
	.endif
	.endif

	.if flash==1
set_addr: pop ZH
	pop ZL
	pop r0
	pop r16
	rcall wr_spi
	pop r16
	rcall wr_spi
	pop r16
	rcall wr_spi
	push r0
	push ZL
	push ZH
	ret

rd_spi: ldi r16,-1
wr_spi: 
	.if xmega==1
	sts (SPIC_DATA),r16
wr_spi_wt: lds r16,(SPIC_STATUS)
	sbrs r16,SPI_IF_bp
	rjmp wr_spi_wt
	lds r16,(SPIC_DATA)
	.else
	out SPDR,r16
wr_spi_wt: in r16,SPSR
	sbrs r16,SPIF
	rjmp wr_spi_wt
	in r16,SPDR
	.endif
	ret

rcv_int: push r0
	.if xmega==1
	lds r0,(SERIAL_IF+USART_DATA_offset)
	st X,r0
	in r0,CPU_SREG
	inc XL
	out CPU_SREG,r0
	.else
	lds r0,UDR0
	st X,r0
	in r0,SREG
	inc XL
	out SREG,r0
	.endif
	pop r0
	reti

.equ rcv_buf=SRAM_START

;q&d mini terminal to store an mp3 file in the SPI flash 
;TeraTerm or ZOC is recommended for transmission
;no transmission protocol, just send the data using "Send file" in BINARY mode
load_flash:
	.if xmega==1
	ldi YL,LOW(SERIAL_IF)
	ldi YH,HIGH(SERIAL_IF)
	ldi r16,LOW(bsel)
	std Y+USART_BAUDCTRLA_offset,r16
	ldi r16,bscale<<4|HIGH(bsel)
	std Y+USART_BAUDCTRLB_offset,r16
	ldi r16,USART_CMODE_ASYNCHRONOUS_gc|USART_CHSIZE_8BIT_gc	;async 8n1
	std Y+USART_CTRLC_offset,r16
	ldi r16,USART_RXEN_bm|USART_TXEN_bm
	std Y+USART_CTRLB_offset,r16
	ldi r16,USART_RXCIF_bm
	std Y+USART_STATUS_offset,r16

	ldi r16,PMIC_HILVLEN_bm|PMIC_MEDLVLEN_bm|PMIC_LOLVLEN_bm	;enable all interrupt levels
	sts (PMIC_CTRL),r16
	.else
	clr r16
	sts UBRR0H,r16
	ldi r16,serial_speed
	sts UBRR0L,r16
	ldi r16,1<<TXC0|1<<U2X0
	sts UCSR0A,r16
	ldi r16,1<<RXEN0|1<<TXEN0
	sts UCSR0B,r16
	ldi r16,1<<USBS0|1<<UCSZ01|1<<UCSZ00	;8n1
	sts UCSR0C,r16
	.endif

	clr YL
	clr YH

erase_lp:
	cbi SPI_SS
	ldi r16,flash_rd_data
	rcall wr_spi
	clr r16
	push r16
	push YL
	push YH
	rcall set_addr

	clr r17
	ldi r18,$10
erase_chk_lp: rcall rd_spi
	cpi r16,-1
	brne erase1
	dec r17
	brne erase_chk_lp
	dec r18
	brne erase_chk_lp
	sbi SPI_SS
	rjmp load_flash1
erase1: 
	sbi SPI_SS

	cbi SPI_SS
	ldi r16,flash_wr_enable
	rcall wr_spi
	sbi SPI_SS
	cbi SPI_SS
	ldi r16,flash_sect_erase
	rcall wr_spi
	clr r16
	push r16
	push YL
	push YH
	rcall set_addr
	sbi SPI_SS
	clr r1
	ldi r16,1+int(clock/16)
erase_delay: dec r1
	brne erase_delay
	dec r16
	brne erase_delay
	cbi SPI_SS
	ldi r16,flash_rd_sr1
	rcall wr_spi
erase_wait: rcall rd_spi
	andi r16,3
	brne erase_wait
	sbi SPI_SS

	mov r16,YH
	rcall disp_hex
	mov r16,YL
	rcall disp_hex
	ldi ZL,LOW(erased_tx<<1)
	ldi ZH,HIGH(erased_tx<<1)
	rcall send_str
	adiw Y,$10
	cpi YH,128
	brne erase_lp

load_flash1: ldi ZL,LOW(send_tx<<1)
	ldi ZH,HIGH(send_tx<<1)
	rcall send_str
	ldi XL,LOW(rcv_buf)
	ldi XH,HIGH(rcv_buf)
	movw Y,X
	clr r17
	clr r18
	.if xmega==1
	lds r0,(SERIAL_IF+USART_DATA_offset)		;dummy read
	lds r16,(SERIAL_IF+USART_CTRLA_offset)
	ori r16,USART_RXCINTLVL_LO_gc				;low rx interrupt priority
	sts (SERIAL_IF+USART_CTRLA_offset),r16
	.else
	lds r0,UDR0			;dummy read
	lds r16,UCSR0B
	ori r16,1<<RXCIE0	;enable rxint
	sts UCSR0B,r16
	.endif
	sei
load_lp1:
	cbi SPI_SS
	ldi r16,flash_wr_enable
	rcall wr_spi
	sbi SPI_SS
	nop
	nop
	nop
	nop
	cbi SPI_SS
	ldi r16,flash_page_pgm
	rcall wr_spi
	clr r16
	push r16
	push r17
	push r18
	rcall set_addr
load_lp: clr r1
	clr r2
	clr r3
	ldi r16,1+int(clock/16)
load_wt: cp XL,YL
	brne load1
	dec r1
	brne load_wt
	dec r2
	brne load_wt
	dec r3
	brne load_wt
	dec r16
	brne load_wt
	sbi SPI_SS
load_end: cli
	.if xmega==1
	lds r16,(SERIAL_IF+USART_CTRLA_offset)
	andi r16,~USART_RXCINTLVL_LO_gc				;disable low rx interrupt priority
	sts (SERIAL_IF+USART_CTRLA_offset),r16
	.else
	lds r16,UCSR0B
	andi r16,~(1<<RXCIE0)						;disable rxint
	sts UCSR0B,r16
	.endif
	ldi ZL,LOW(finish_tx<<1)
	ldi ZH,HIGH(finish_tx<<1)
	rcall send_str
load_sleep:	sleep
	rjmp load_sleep
load1:
	ld r16,Y
	rcall wr_spi
	inc YL
	brne load_lp
	sbi SPI_SS
	clr r24
	ldi r25,10
load_delay: sbiw r25:r24,1
	brne load_delay 
	cbi SPI_SS
	ldi r16,flash_rd_sr1
	rcall wr_spi
load_wt1: rcall rd_spi
	andi r16,1
	brne load_wt1
	sbi SPI_SS
	inc r17
	brne load_lp1
	inc r18
	cpi r18,128
	breq load_end
	rjmp load_lp1

	.if flash==1
disp_hex: push r16
	swap r16
	rcall disp_nibble
	pop r16
disp_nibble: andi r16,15
	subi r16,-'0'
	cpi r16,$3a
	brcs disp_nibble1
	subi r16,-7
disp_nibble1: push r16
	.endif
	.if xmega==1
disp_chr:
	lds r16,(SERIAL_IF+USART_STATUS_offset)
	sbrs r16,USART_DREIF_bp
	rjmp disp_chr
	pop r16
	sts (SERIAL_IF+USART_DATA_offset),r16
	.else
disp_chr:
	lds r16,UCSR0A
	sbrs r16,UDRE0
	rjmp disp_chr
	pop r16
	sts UDR0,r16
	.endif
	ret

send_str: lpm r16,Z
	andi r16,$7f
	push r16
	.if xmega==1
send_chr:
	lds r16,(SERIAL_IF+USART_STATUS_offset)
	sbrs r16,USART_DREIF_bp
	rjmp send_chr
	pop r16
	sts (SERIAL_IF+USART_DATA_offset),r16
	.else
send_chr:
	lds r16,UCSR0A
	sbrs r16,UDRE0
	rjmp send_chr
	pop r16
	sts UDR0,r16
	.endif
	lpm r16,Z+
	or r16,r16
	brpl send_str
	ret

send_tx: .db 13,10,"Start to send data in BINARY format..."
nl_tx: .db 13,10+$80
finish_tx: .db 13,10,"Loading finished.",13,10+$80			,-1
erased_tx: .db " erased",13,10+$80							,-1
	.endif

	.if chip==1 && speedcheck==1
disp_hex: push r16
	swap r16
	rcall disp_nibble
	pop r16
disp_nibble: andi r16,15
	subi r16,-'0'
	cpi r16,$3a
	brcs disp_nibble1
	subi r16,-7
disp_nibble1: push r16
	.if xmega==1
disp_chr:
	lds r16,(SERIAL_IF+USART_STATUS_offset)
	sbrs r16,USART_DREIF_bp
	rjmp disp_chr
	pop r16
	sts (SERIAL_IF+USART_DATA_offset),r16
	.else
disp_chr:
	lds r16,UCSR0A
	sbrs r16,UDRE0
	rjmp disp_chr
	pop r16
	sts UDR0,r16
	.endif
	ret
	.endif

decode_frame:
	.if flash==1 && br_support==0
	clr ZL
	clr ZH
	.endif
	push ZL
	push ZH
	rcall calc_slen
	pop YH
	pop YL
	ldi ZL,LOW(nrtab_l*2)
	ldi ZH,HIGH(nrtab_l*2)
	.if eval_hd==1
	sbis btype,bt2				;skip if Block_type 2
	rjmp dec_frm_nrmblk
	ldi ZL,LOW(nrtab_s*2)
	ldi ZH,HIGH(nrtab_s*2)
dec_frm_nrmblk:	
	.endif
	add ZL,r16
	adc ZH,r15
	ldi XL,LOW(slen+4)
	ldi XH,HIGH(slen+4)
	ldi r17,4
get_sfi: ld r16,-X
	push r16
	sbiw Z,1
	lpm r16,Z
	push r16
	dec r17
	brne get_sfi

	movw Z,Y
	rd_bbyte
	ldi r18,8
	ldi XL,LOW(scfac)
	ldi XH,HIGH(scfac)
	ldi r17,4
get_sfv: pop r2				;nrtab entry
	pop r3					;slen entry
	tst r2
	breq get_sfvn
get_sfv1: clr r5
	tst r3
	breq get_sfvn1
	mov r4,r3
get_sfv2: lsl r1
	rol r0
	rol r5
	dec r18
	brne get_sfv3
	rd_bbyte
	ldi r18,8
get_sfv3: dec r4
	brne get_sfv2
get_sfvn1: st X+,r5
	dec r2
	brne get_sfv1
get_sfvn: dec r17
	brne get_sfv
clr_sfv: cpi XL,LOW(scfac+36)
	breq clr_sfvx
	st X+,r15
	rjmp clr_sfv
clr_sfvx:

	push YH					;save start of part2_3
	push YL
	ldi XL,LOW(frame)
	ldi XH,HIGH(frame)
	movw Y,Z
	lds r16,(bigv)
	lds r17,(bigv+1)
	mov r2,r16
	or r2,r17
	brne dec_bigv
	rjmp dec_skip_bigv
dec_bigv:
	.if eval_hd==1
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	.else
	ldi ZL,LOW(sfbl*2)
	ldi ZH,HIGH(sfbl*2)
	.endif
	lds r24,(r0cnt)
	add r24,r24
	add ZL,r24
	adc ZH,r15
	lpm r24,Z+
	lpm r25,Z
	.if eval_hd==1
	sbis btype,altwnd		;check if alternate window
	rjmp dec_frm1
	cp r16,r24
	cpc r17,r25
	brcc dec_frm1
	movw r25:r24,r17:r16
dec_frm1:
	.endif
	push r25
	push r24
	lds r16,(table1)
	rcall huff_dec
	.if eval_hd==1
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	.else
	ldi ZL,LOW(sfbl*2)
	ldi ZH,HIGH(sfbl*2)
	.endif
	lds r16,(r0cnt)
	lds r24,(r1cnt)
	add r24,r16
	inc r24
	add r24,r24
	add ZL,r24
	adc ZH,r15
	lpm r24,Z+
	lpm r25,Z

	.if eval_hd==1
	sbis btype,altwnd		;check if alternate window
	rjmp dec_frm2
	lds r16,(bigv)
	lds r17,(bigv+1)
	mov r2,r16
	or r2,r17
	breq dec_frm2
	cp r16,r24
	cpc r17,r25
	brcc dec_frm2
	movw r25:r24,r17:r16
dec_frm2:
	.endif
	pop r16
	pop r17
	push r25
	push r24
	sub r24,r16
	sbc r25,r17
	lds r16,(table2)
	rcall huff_dec
	lds r24,(bigv)
	lds r25,(bigv+1)
	pop r16
	pop r17
	sub r24,r16
	sbc r25,r17
	brpl dec_frm3
	pop r16
	pop r16
	rjmp fill_rzero
dec_frm3: lds r16,(table3)
	rcall huff_dec

dec_skip_bigv: movw r17:r16,Y	;calculate bits already used by region 0...2
	pop r24
	pop r25
	sub r16,r24
	sbc r17,r25
	.if br_support==1
	brcc dec_frm4
	subi r16,LOW(-fr_buf_size)
	sbci r17,HIGH(-fr_buf_size)
dec_frm4:
	.endif
	subi r16,1
	sbci r17,0
	add r16,r16
	adc r17,r17
	add r16,r16
	adc r17,r17
	add r16,r16
	adc r17,r17
	ldi r24,8
	sub r24,r18
	add r16,r24
	adc r17,r15

	lds r24,(part23len)
	lds r25,(part23len+1)
	sub r24,r16
	sbc r25,r17
	brcs fill_rzero
	mov r16,r24
	or r16,r25
	breq fill_rzero
	lds r16,(cnt1tbl)
	push r16
	sbrs r16,0
	rcall hufftbla
	pop r16
	sbrc r16,0
	rcall hufftblb

fill_rzero:	ldi r17,HIGH(frame+2*576)
clr_high: cpi XL,LOW(frame+2*576)
	cpc XH,r17
	brcc decode_x
	st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	rjmp clr_high
decode_x: ret

huff_fill: st X+,r15
	st X+,r15
	st X+,r15
	st X+,r15
	sbiw r25:r24,1
	brne huff_fill
huff_decx: ret

huff_shft1: rd_bbyte
	ldi r18,8
	rjmp huff_shft2

;huffman decoder is capable to handle up to 4 data bits at once
;the decoding tables are built as follows:
;
;always 4 bits are used as an index to a 16 byte sub table
;this sub table has two different types of entries
;both types carry a four bit forward-relative pointer to a second sub table in the lower nibble
;
;a: Bit 7:	 0
;   Bit 6:   0 - entry is a pointer to next sub table
;            1 - entry is a pointer to special one-bit table
;   Bit 5-4: upper two bits from pointer
;   the entry in second sub table carries the lower eight bits of the pointer
;	the pointer is coded as forward-relative value so a maximum distance of 1023 words can be skipped
;
;b: Bit 7:   1
;   Bit 6:   unused, always 0
;   Bit 5-4: data bit count minus one for this entry (so range 0...3 means one to four data bits)
;   the entry in second sub table carries the decoded value
;
;if only a single bit of a particular code word remains for decoding, a special one-bit table is used
;containing just the two value pairs

huff_dec:
	mov r17,r24
	or r17,r25
	breq huff_decx
	tst r16
	breq huff_fill
	ldi ZL,LOW(htables*2)
	ldi ZH,HIGH(htables*2)
	add r16,r16
	add ZL,r16
	adc ZH,r15
	lpm r2,Z+
	lpm r16,Z+
	mov r17,r16
	andi r16,15
	add r2,r2
	adc r16,r16
	add ZL,r2
	adc ZH,r16
	movw r3:r2,Z
	swap r17
	clt
	andi r17,15
	breq huffx1
	set
huffx1:	mov r4,r17		;linbits
hufflp: movw Z,r3:r2
hufflp1: mov r16,r0
	swap r16
	andi r16,15
	add ZL,r16
	adc ZH,r15
	lpm r16,Z+
	sbrs r16,7
	rjmp huff_brtbl
	ldi r17,15
	and r17,r16
	add ZL,r17
	adc ZH,r15
huff_single: lpm r17,Z
	movw Z,Y
	swap r16
	andi r16,3
huff_shft: lsl r1
	rol r0
	dec r18
	breq huff_shft1
huff_shft2: dec r16
	brpl huff_shft
	mov r16,r17
	swap r16
	clr r19
	andi r16,15
	breq huff_x2
	brtc huff_nolin1
	cpi r16,15
	brne huff_nolin1
	rcall huff_linb
huff_nolin1: lsl r1
	rol r0
	brcc huff_x1
	com r19
	neg r16
	sbci r19,-1
huff_x1: dec r18
	breq huff_x
huff_x2: st X+,r16
	st X+,r19
	clr r19
	mov r16,r17
	andi r16,15
	breq huff_y2
	brtc huff_nolin2
	cpi r16,15
	brne huff_nolin2
	rcall huff_linb
huff_nolin2: lsl r1
	rol r0
	brcc huff_y1
	com r19
	neg r16
	sbci r19,-1
huff_y1: dec r18
	breq huff_y
huff_y2: st X+,r16
	st X+,r19
	movw Y,Z
	sbiw r25:r24,1
	breq huff_decxx
	rjmp hufflp
huff_decxx: ret

huff_x: rd_bbyte
	ldi r18,8
	rjmp huff_x2

huff_brtbl: mov r17,r16
	swap r17
	andi r16,15
	add ZL,r16
	adc ZH,r15
	lpm r16,Z
	add ZL,r16
	ldi r16,3
	and r16,r17
	adc ZH,r16
	lsl r1
	rol r0
	dec r18
	breq huff_shft11
huff_shft12: 
	lsl r1
	rol r0
	dec r18
	breq huff_shft21
huff_shft22: 
	lsl r1
	rol r0
	dec r18
	breq huff_shft31
huff_shft32: 
	lsl r1
	rol r0
	dec r18
	breq huff_shft41
huff_shft42: 
	sbrs r17,2		;special handling for 1 bit table
	rjmp hufflp1
	sbrc r0,7
	adiw Z,1
	clr r16
	rjmp huff_single


huff_y:	rd_bbyte
	ldi r18,8
	rjmp huff_y2

huff_shft11: movw r23:r22,Z
	movw Z,Y
	rd_bbyte
	movw Y,Z
	movw Z,r23:r22
	ldi r18,8
	rjmp huff_shft12
huff_shft21: movw r23:r22,Z
	movw Z,Y
	rd_bbyte
	movw Y,Z
	movw Z,r23:r22
	ldi r18,8
	rjmp huff_shft22
huff_shft31: movw r23:r22,Z
	movw Z,Y
	rd_bbyte
	movw Y,Z
	movw Z,r23:r22
	ldi r18,8
	rjmp huff_shft32
huff_shft41: movw r23:r22,Z
	movw Z,Y
	rd_bbyte
	movw Y,Z
	movw Z,r23:r22
	ldi r18,8
	rjmp huff_shft42

huff_linb: clr r16
	mov r5,r4
huff_linblp: lsl r1
	rol r0
	rol r16
	rol r19
	dec r18
	breq huff_linb1
huff_linb2: dec r5
	brne huff_linblp
	subi r16,-15
	sbci r19,-1
	ret
huff_linb1: rd_bbyte
	ldi r18,8
	rjmp huff_linb2

huffa_shft1: movw Z,Y
	rd_bbyte
	movw Y,Z
	ldi r18,8
	rjmp huffa_shft2

hufftbla:
huffalp: ldi ZL,LOW(tbla*2)
	ldi ZH,HIGH(tbla*2)
	mov r16,r0
	lsr r16
	lsr r16
	add ZL,r16
	adc ZH,r15
	lpm r17,Z
	mov r16,r17
	swap r16
	andi r16,15
huffa_shft: lsl r1
	rol r0
	dec r18
	breq huffa_shft1
huffa_shft2: sbiw r25:r24,1
	dec r16
	brpl huffa_shft
	ldi ZL,LOW(tblabv*2)
	ldi ZH,HIGH(tblabv*2)
	andi r17,15
	lsl r17
	lsl r17
	add ZL,r17
	adc ZH,r15
	lpm r2,Z+
	lpm r3,Z+
	lpm r4,Z+
	lpm r5,Z
	movw Z,Y
	clr r16
	tst r2
	breq huffa_v2
	lsl r1
	rol r0
	brcc huffa_v1
	neg r2
	com r16
huffa_v1: sbiw r25:r24,1
	dec r18
	breq huffa_v
huffa_v2: st X+,r2
	st X+,r16
	clr r16
	tst r3
	breq huffa_w2
	lsl r1
	rol r0
	brcc huffa_w1
	neg r3
	com r16
huffa_w1: sbiw r25:r24,1
	dec r18
	breq huffa_w
huffa_w2: st X+,r3
	st X+,r16
	clr r16
	tst r4
	breq huffa_x2
	lsl r1
	rol r0
	brcc huffa_x1
	neg r4
	com r16
huffa_x1: sbiw r25:r24,1
	dec r18
	breq huffa_x
huffa_x2: st X+,r4
	st X+,r16
	clr r16
	tst r5
	breq huffa_y2
	lsl r1
	rol r0
	brcc huffa_y1
	neg r5
	com r16
huffa_y1: sbiw r25:r24,1
	dec r18
	breq huffa_y
huffa_y2: st X+,r5
	st X+,r16
	movw Y,Z
	tst r25
	brmi huffa_xx
	mov r16,r24
	or r16,r25
	breq huffa_xx
	rjmp huffalp
huffa_xx: ret

huffa_v: rd_bbyte
	ldi r18,8
	rjmp huffa_v2
huffa_w: rd_bbyte
	ldi r18,8
	rjmp huffa_w2
huffa_x: rd_bbyte
	ldi r18,8
	rjmp huffa_x2
huffa_y: rd_bbyte
	ldi r18,8
	rjmp huffa_y2

huffb_shft11: rd_bbyte
	ldi r18,8
	rjmp huffb_shft12
huffb_shft21: rd_bbyte
	ldi r18,8
	rjmp huffb_shft22
huffb_shft31: rd_bbyte
	ldi r18,8
	rjmp huffb_shft32
huffb_shft41: rd_bbyte
	ldi r18,8
	rjmp huffb_shft42

hufftblb:
huffblp: ldi ZL,LOW(tblabv*2)
	ldi ZH,HIGH(tblabv*2)
	mov r16,r0
	com r16
	lsr r16
	lsr r16
	andi r16,0b00111100
	add ZL,r16
	adc ZH,r15
	lpm r2,Z+
	lpm r3,Z+
	lpm r4,Z+
	lpm r5,Z
	movw Z,Y
	sbiw r25:r24,4
	lsl r1
	rol r0
	dec r18
	breq huffb_shft11
huffb_shft12: lsl r1
	rol r0
	dec r18
	breq huffb_shft21
huffb_shft22: lsl r1
	rol r0
	dec r18
	breq huffb_shft31
huffb_shft32: lsl r1
	rol r0
	dec r18
	breq huffb_shft41
huffb_shft42:
	clr r16
	tst r2
	breq huffb_v2
	lsl r1
	rol r0
	brcc huffb_v1
	neg r2
	com r16
huffb_v1: sbiw r25:r24,1
	dec r18
	breq huffb_v
huffb_v2: st X+,r2
	st X+,r16
	clr r16
	tst r3
	breq huffb_w2
	lsl r1
	rol r0
	brcc huffb_w1
	neg r3
	com r16
huffb_w1: sbiw r25:r24,1
	dec r18
	breq huffb_w
huffb_w2: st X+,r3
	st X+,r16
	clr r16
	tst r4
	breq huffb_x2
	lsl r1
	rol r0
	brcc huffb_x1
	neg r4
	com r16
huffb_x1: sbiw r25:r24,1
	dec r18
	breq huffb_x
huffb_x2: st X+,r4
	st X+,r16
	clr r16
	tst r5
	breq huffb_y2
	lsl r1
	rol r0
	brcc huffb_y1
	neg r5
	com r16
huffb_y1: sbiw r25:r24,1
	dec r18
	breq huffb_y
huffb_y2: st X+,r5
	st X+,r16
	movw Y,Z
	tst r25
	brmi huffb_xx
	mov r16,r24
	or r16,r25
	breq huffb_xx
	rjmp huffblp
huffb_xx: ret

huffb_v: rd_bbyte
	ldi r18,8
	rjmp huffb_v2
huffb_w: rd_bbyte
	ldi r18,8
	rjmp huffb_w2
huffb_x: rd_bbyte
	ldi r18,8
	rjmp huffb_x2
huffb_y: rd_bbyte
	ldi r18,8
	rjmp huffb_y2

;reference implementation of slen calculation taken from Helix decoder
calc_slen: ldi YL,LOW(slen)
	ldi YH,HIGH(slen)
	ldi ZL,LOW(div5tab*2)
	ldi ZH,HIGH(div5tab*2)
	lds r16,(scfacc)
	lds r17,(scfacc+1)
	ldi r18,HIGH(400)
	cpi r16,LOW(400)
	cpc r17,r18
	brcc calc_slen1
	ldi r18,3
	and r18,r16
	std Y+3,r18		;slen3
	lsr r17
	ror r16
	lsr r16
	ldi r18,3
	and r18,r16
	std Y+2,r18		;slen2
	lsr r16
	lsr r16
	add ZL,r16
	adc ZH,r15
	lpm r16,Z
	mov r17,r16
	swap r16
	andi r16,15
	st Y,r16		;slen0   /5
	andi r17,15
	std Y+1,r17		;slen1   %5
	ldi r16,4		;nrtab index
	ret
calc_slen1: ldi r18,HIGH(500)
	cpi r16,LOW(500)
	cpc r17,r18
	brcc calc_slen2
	subi r16,LOW(400)
	ldi r18,3
	and r18,r16
	std Y+3,r18		;slen2
	lsr r16
	lsr r16
	add ZL,r16
	adc ZH,r15
	std Y+3,r15		;slen3=0
	lpm r16,Z
	mov r17,r16
	swap r16
	andi r16,15
	st Y,r16		;slen0   /5
	andi r17,15
	std Y+1,r17		;slen1   %5
	ldi r16,8		;nrtab index
	ret
calc_slen2: subi r16,LOW(500)
	ldi ZL,LOW(div3tab*2)
	ldi ZH,HIGH(div3tab*2)
	subi r16,LOW(500)
	add ZL,r16
	adc ZH,r15
	std Y+2,r15		;slen2=0
	std Y+3,r15		;slen3=0
	lpm r16,Z
	mov r17,r16
	swap r16
	andi r16,15
	st Y,r16		;slen0   /3
	andi r17,15
	std Y+1,r17		;slen1   %3
	ldi r16,12		;nrtab index
	ret

;requantization implementation based on libmad 0.15
;requantization is performed using 16 bit values for the power of 4/3
;since 16 bit cannot provide power values for source range 0...8207 and to retain accuracy for lower source values
;the source data is appropriately shifted if exceeding certain thresholds to keep the result in applicable range 
	.if reqmode==1
	.macro reqval
	ld ZL,Y
	ldd ZH,Y+1
	mov r0,ZL
	or r0,ZH
	brne req0
	adiw Y,2
	rjmp req_skip1
req0: mov r23,ZH
	tst ZH
	brpl req1
	com ZH
	neg ZL
	sbci ZH,-1
req1: movw r21:r20,r3:r2
	mov r22,r4
	add ZL,ZL
	adc ZH,ZH
	breq req_s10
	cpi ZH,HIGH(4096*2)
	brcc req_s11
	cpi ZH,HIGH(2048*2)
	brcc req_s12
	cpi ZH,HIGH(1024*2)
	brcc req_s13
	cpi ZH,HIGH(512*2)
	brcc req_s14
	cpi ZH,HIGH(256*2)
	brcc req_s15
	rjmp req_s16
req_s10: tst ZL
	brpl req_s18
	rjmp req_s17
req_s11: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s12: add r20,r20
	adc r21,r21
	adc r22,r22
req_s13: add r20,r20
	adc r21,r21
	adc r22,r22
req_s14: add r20,r20
	adc r21,r21
	adc r22,r22
req_s15: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s16: add r20,r20
	adc r21,r21
	adc r22,r22
req_s17: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s18:
	subi ZL,LOW(-(powtab*2))
	sbci ZH,HIGH(-(powtab*2))
	lpm r16,Z+
	lpm r17,Z

	mul	r17,r21
	movw Z,r1:r0
	mul	r16,r20
	mov r5,r1
	mul	r16,r21
	add	r5,r0
	adc	ZL,r1
	adc	ZH,r15
	mul	r17,r20
	add	r5,r0
	adc	ZL,r1
	adc	ZH,r15
	mul r16,r22
	add ZL,r0
	adc ZH,r1
	mul r17,r22
	add ZH,r0

	add r5,r5
	adc ZL,ZL
	adc ZH,ZH

	.if round==1
	add r4,r4
	adc ZL,r15
	adc ZH,r15
	.endif

	tst r23
	brpl req2
	com ZH
	neg ZL
	sbci ZH,-1
req2: st Y+,ZL
	st Y+,ZH
req_skip1: 
	.endmacro
	
	.macro exp_calc
	add r8,r8
	movw Z,r25:r24
	add ZL,r8
	adc ZH,r15
	lpm r4,Z+
	lpm r5,Z
	mul r22,r4
	mov r7,r1
	mul r23,r5
	movw r9:r8,r1:r0
	mul r22,r5
	add r7,r0
	adc r8,r1
	adc r9,r15
	mul r23,r4
	add r7,r0
	adc r8,r1
	adc r9,r15

	brtc exp_calc1
	mov r7,r9
	clr r8
	clr r9
	lsr r7
	rjmp exp_calc2
exp_calc1: lsr r9
	ror r8
	ror r7
	lsr r9
	ror r8
	ror r7
exp_calc2: 
	.if round==1
	adc r7,r15
	adc r8,r15
	adc r9,r15
	.endif
	.endmacro

	.macro req_calc
	clr r22
	clr r23
	clt
	subi r16,132
	brcs req_calc5
req_calc1: cpi r16,210-132
	brcs req_calc2
	lds r22,(max_gain)
	lds r23,(max_gain+1)
	rjmp req_calc4
req_calc2: cpi r16,152-132
	brcc req_calc3
	set
req_calc3: add r16,r16
	ldi ZL,LOW(reqtabgain*2)
	ldi ZH,HIGH(reqtabgain*2)
	add ZL,r16
	adc ZH,r15
	lpm r22,Z+
	lpm r23,Z
req_calc4:
	ldi r24,LOW(reqtabscfac*2)
	ldi r25,HIGH(reqtabscfac*2)
	lds r16,(scfacsc)
	sbrc r16,0
	adiw r25:r24,32
req_calc5:
	.endmacro

	.else

reqval_: ld ZL,Y
	ldd ZH,Y+1
	mov r0,ZL
	or r0,ZH
	brne req0
	adiw Y,2
	ret
req0: mov r23,ZH
	tst ZH
	brpl req1
	com ZH
	neg ZL
	sbci ZH,-1
req1: movw r21:r20,r3:r2
	mov r22,r4
	add ZL,ZL
	adc ZH,ZH
	breq req_s10
	cpi ZH,HIGH(4096*2)
	brcc req_s11
	cpi ZH,HIGH(2048*2)
	brcc req_s12
	cpi ZH,HIGH(1024*2)
	brcc req_s13
	cpi ZH,HIGH(512*2)
	brcc req_s14
	cpi ZH,HIGH(256*2)
	brcc req_s15
	rjmp req_s16
req_s10: tst ZL
	brpl req_s18
	rjmp req_s17
req_s11: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s12: add r20,r20
	adc r21,r21
	adc r22,r22
req_s13: add r20,r20
	adc r21,r21
	adc r22,r22
req_s14: add r20,r20
	adc r21,r21
	adc r22,r22
req_s15: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s16: add r20,r20
	adc r21,r21
	adc r22,r22
req_s17: add r20,r20
	adc r21,r21
	adc r22,r22
	add r20,r20
	adc r21,r21
	adc r22,r22
req_s18:
	subi ZL,LOW(-(powtab*2))
	sbci ZH,HIGH(-(powtab*2))
	lpm r16,Z+
	lpm r17,Z

	mul	r17,r21
	movw Z,r1:r0
	mul	r16,r20
	mov r5,r1
	mul	r16,r21
	add	r5,r0
	adc	ZL,r1
	adc	ZH,r15
	mul	r17,r20
	add	r5,r0
	adc	ZL,r1
	adc	ZH,r15
	mul r16,r22
	add ZL,r0
	adc ZH,r1
	mul r17,r22
	add ZH,r0

	add r5,r5
	adc ZL,ZL
	adc ZH,ZH

	.if round==1
	add r4,r4
	adc ZL,r15
	adc ZH,r15
	.endif

	tst r23
	brpl req2
	com ZH
	neg ZL
	sbci ZH,-1
req2: st Y+,ZL
	st Y+,ZH
	ret
	
exp_calc_: add r8,r8
	movw Z,r25:r24
	add ZL,r8
	adc ZH,r15
	lpm r4,Z+
	lpm r5,Z
	mul r22,r4
	mov r7,r1
	mul r23,r5
	movw r9:r8,r1:r0
	mul r22,r5
	add r7,r0
	adc r8,r1
	adc r9,r15
	mul r23,r4
	add r7,r0
	adc r8,r1
	adc r9,r15

	brtc exp_calc1
	mov r7,r9
	clr r8
	clr r9
	lsr r7
	rjmp exp_calc2
exp_calc1: lsr r9
	ror r8
	ror r7
	lsr r9
	ror r8
	ror r7
exp_calc2: 
	.if round==1
	adc r7,r15
	adc r8,r15
	adc r9,r15
	.endif
	ret

req_calc_: clr r22
	clr r23
	clt
	subi r16,132
	brcs req_calc5
req_calc1: cpi r16,210-132
	brcs req_calc2
	lds r22,(max_gain)
	lds r23,(max_gain+1)
	rjmp req_calc4
req_calc2: cpi r16,152-132
	brcc req_calc3
	set
req_calc3: add r16,r16
	ldi ZL,LOW(reqtabgain*2)
	ldi ZH,HIGH(reqtabgain*2)
	add ZL,r16
	adc ZH,r15
	lpm r22,Z+
	lpm r23,Z
req_calc4:
	ldi r24,LOW(reqtabscfac*2)
	ldi r25,HIGH(reqtabscfac*2)
	lds r16,(scfacsc)
	sbrc r16,0
	adiw r25:r24,32
req_calc5: ret

	.macro reqval
	rcall reqval_
	.endmacro
	.macro exp_calc
	rcall exp_calc_
	.endmacro
	.macro req_calc
	rcall req_calc_
	.endmacro
	.endif

requantize:
	.if eval_hd==1
	sbis btype,bt2		;skip if Block_type 2
	rjmp requantize_l
	rcall requantize_s
	rjmp req_c
requantize_l:
	.endif
	lds r16,(ggain)
	req_calc
	ldi XL,LOW(scfac)
	ldi XH,HIGH(scfac)
	ldi YL,LOW(scfac1)
	ldi YH,HIGH(scfac1)
	ldi r17,22
exp_calc_lp: ld r8,X+
	exp_calc
	st Y+,r7
	st Y+,r8
	st Y+,r9
	dec r17
	brne exp_calc_lp

req_c: ldi XL,LOW(scfac1)
	ldi XH,HIGH(scfac1)
	ldi YL,LOW(frame)
	ldi YH,HIGH(frame)
	clr r24
	clr r25
	ldi r18,22
	.if eval_hd==1
	sbic btype,bt2			;skip if not Block_type 2
	ldi r18,39
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	.else
	ldi ZL,LOW(sfbl*2)
	ldi ZH,HIGH(sfbl*2)
	.endif
req_lp2: lpm r10,Z+
	lpm r11,Z+
	movw r13:r12,Z
	ld r2,X+
	ld r3,X+
	ld r4,X+
req_lp1:
	reqval
	reqval
	adiw r25:r24,1
	cp r24,r10
	cpc r25,r11
	breq req5
	rjmp req_lp1
req5: movw Z,r13:r12
	dec r18
	breq reqx
	rjmp req_lp2
reqx: ret

	.if eval_hd==1
requantize_s:
	lds r17,(sblk_gain)
	ldi r18,0
	rcall req_sblk
	lds r17,(sblk_gain+1)
	ldi r18,1
	rcall req_sblk
	lds r17,(sblk_gain+2)
	ldi r18,2
req_sblk: 
	add r17,r17
	add r17,r17
	add r17,r17
	lds r16,(ggain)
	sub r16,r17
	req_calc
	ldi XL,LOW(scfac)
	ldi XH,HIGH(scfac)
	add XL,r18
	adc XH,r15
	ldi YL,LOW(scfac1)
	ldi YH,HIGH(scfac1)
	ldi r17,3
	mul r17,r18
	add YL,r0
	adc YH,r1
	ldi r17,13
exp_scalc_lp: ld r8,X
	exp_calc
	st Y,r7
	std Y+1,r8
	std Y+2,r9
	adiw X,3
	adiw Y,9
	dec r17
	brne exp_scalc_lp
	ret

;reorder data in case of short blocks
reorder:
	ldi XL,LOW(output)
	ldi XH,HIGH(output)
	sbis btype,cbuf
	ldi XL,LOW(output1)
	sbis btype,cbuf
	ldi XH,HIGH(output1)
	movw r21:r20,X
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	ldi YL,LOW(frame)
	ldi YH,HIGH(frame)
	rcall reorder_sblk
	movw X,r21:r20
	adiw X,6*2
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	lpm YL,Z+
	lpm YH,Z+
	add YL,YL
	adc YH,YH
	add YL,YL
	adc YH,YH
	subi YL,LOW(-frame)
	sbci YH,HIGH(-frame)
	rcall reorder_sblk
	movw X,r21:r20
	adiw X,12*2
	lds ZL,(sfb_ptr)
	lds ZH,(sfb_ptr+1)
	adiw Z,2
	lpm YL,Z+
	lpm YH,Z+
	add YL,YL
	adc YH,YH
	add YL,YL
	adc YH,YH
	subi YL,LOW(-frame)
	sbci YH,HIGH(-frame)
	rcall reorder_sblk
	ldi r16,576/4
	movw X,r21:r20
	ldi YL,LOW(frame)
	ldi YH,HIGH(frame)
reorder_cp_lp:
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	ld r0,X+
	st Y+,r0
	dec r16
	brne reorder_cp_lp
	ret

reorder_sblk:
	ldi r19,13
	ldi r18,6
reorder_sblk_lp: lpm r16,Z+
	lpm r17,Z+
	add r16,r16
	adc r17,r17
	add r16,r16
	adc r17,r17
	subi r16,LOW(-frame)
	sbci r17,HIGH(-frame)
	dec r19
	brpl reorder_sblk1
	ret
reorder_sblk1:
	ld r0,Y+
	ld r1,Y+
	st X+,r0
	st X+,r1
	dec r18
	brne reorder_sblk2
	adiw X,24
	ldi r18,6
reorder_sblk2:
	cp r16,YL
	cpc r17,YH
	brne reorder_sblk1
	adiw Z,2
	lpm YL,Z+
	lpm YH,Z+
	add YL,YL
	adc YH,YH
	add YL,YL
	adc YH,YH
	subi YL,LOW(-frame)
	sbci YH,HIGH(-frame)
	rjmp reorder_sblk_lp
	.endif

	.if aalmode==1

	.macro bca_acs
	fmuls r19,r23		;b*ca
	movw r5:r4,r1:r0
	fmul r18,r22
	adc r4,r15
	mov r3,r1
	fmulsu r19,r22
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	fmulsu r23,r18
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.if round==1
	add r3,r3
	adc r4,r15
	adc r5,r15
	.endif

	fmuls r17,r21		;a*cs
	movw r7:r6,r1:r0
	fmul r16,r20
	adc r6,r15
	mov r3,r1
	fmulsu r17,r20
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	fmulsu r21,r16
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	.if round==1
	add r3,r3
	adc r6,r15
	adc r7,r15
	.endif
	sub r6,r4
	sbc r7,r5
	.endmacro

	.macro aca_bcs
	fmuls r17,r23		;a*ca
	movw r5:r4,r1:r0
	fmul r16,r22
	adc r4,r15
	mov r3,r1
	fmulsu r17,r22
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	fmulsu r23,r16
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.if round==1
	add r3,r3
	adc r4,r15
	adc r5,r15
	.endif

	fmuls r19,r21		;b*cs
	movw r7:r6,r1:r0
	fmul r18,r20
	adc r6,r15
	mov r3,r1
	fmulsu r19,r20
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	fmulsu r21,r18
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	.if round==1
	add r3,r3
	adc r6,r15
	adc r7,r15
	.endif
	add r6,r4
	adc r7,r5
	.endmacro

	.else

	.macro bca_acs
	rcall bca_acs_
	.endmacro

	.macro aca_bcs
	rcall aca_bcs_
	.endmacro

bca_acs_: fmuls r19,r23		;b*ca
	movw r5:r4,r1:r0
	fmul r18,r22
	adc r4,r15
	mov r3,r1
	fmulsu r19,r22
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	fmulsu r23,r18
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.if round==1
	add r3,r3
	adc r4,r15
	adc r5,r15
	.endif

	fmuls r17,r21		;a*cs
	movw r7:r6,r1:r0
	fmul r16,r20
	adc r6,r15
	mov r3,r1
	fmulsu r17,r20
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	fmulsu r21,r16
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	.if round==1
	add r3,r3
	adc r6,r15
	adc r7,r15
	.endif
	sub r6,r4
	sbc r7,r5
	ret

aca_bcs_: fmuls r17,r23		;a*ca
	movw r5:r4,r1:r0
	fmul r16,r22
	adc r4,r15
	mov r3,r1
	fmulsu r17,r22
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	fmulsu r23,r16
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.if round==1
	add r3,r3
	adc r4,r15
	adc r5,r15
	.endif

	fmuls r19,r21		;b*cs
	movw r7:r6,r1:r0
	fmul r18,r20
	adc r6,r15
	mov r3,r1
	fmulsu r19,r20
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	fmulsu r21,r18
	sbc r7,r15
	add r3,r0
	adc r6,r1
	adc r7,r15
	.if round==1
	add r3,r3
	adc r6,r15
	adc r7,r15
	.endif
	add r6,r4
	adc r7,r5
	ret

	.endif

;reference implementation of antialiasing taken from libmad 0.15
antialiase: ldi r24,31
	ldi YL,LOW(frame+10*2)
	ldi YH,HIGH(frame+10*2)
aali_lp: 
	ldi ZL,LOW(csa*2)
	ldi ZH,HIGH(csa*2)

	ldd r16,Y+14	;a
	ldd r17,Y+15
	ldd r18,Y+8*2	;b
	ldd r19,Y+8*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali1
	rjmp aali_skip1
aali1: bca_acs
	std Y+14,r6
	std Y+15,r7
	aca_bcs
	std Y+8*2,r6
	std Y+8*2+1,r7
aali_skip1:

	ldd r16,Y+12	;a
	ldd r17,Y+13
	ldd r18,Y+9*2	;b
	ldd r19,Y+9*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali2
	rjmp aali_skip2
aali2: bca_acs
	std Y+12,r6
	std Y+13,r7
	aca_bcs
	std Y+9*2,r6
	std Y+9*2+1,r7
aali_skip2:

	ldd r16,Y+10	;a
	ldd r17,Y+11
	ldd r18,Y+10*2	;b
	ldd r19,Y+10*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali3
	rjmp aali_skip3
aali3: bca_acs
	std Y+10,r6
	std Y+11,r7
	aca_bcs
	std Y+10*2,r6
	std Y+10*2+1,r7
aali_skip3:

	ldd r16,Y+8	;a
	ldd r17,Y+9
	ldd r18,Y+11*2	;b
	ldd r19,Y+11*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali4
	rjmp aali_skip4
aali4: bca_acs
	std Y+8,r6
	std Y+9,r7
	aca_bcs
	std Y+11*2,r6
	std Y+11*2+1,r7
aali_skip4:

	ldd r16,Y+6	;a
	ldd r17,Y+7
	ldd r18,Y+12*2	;b
	ldd r19,Y+12*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali5
	rjmp aali_skip5
aali5: bca_acs
	std Y+6,r6
	std Y+7,r7
	aca_bcs
	std Y+12*2,r6
	std Y+12*2+1,r7
aali_skip5:

	ldd r16,Y+4	;a
	ldd r17,Y+5
	ldd r18,Y+13*2	;b
	ldd r19,Y+13*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali6
	rjmp aali_skip6
aali6: bca_acs
	std Y+4,r6
	std Y+5,r7
	aca_bcs
	std Y+13*2,r6
	std Y+13*2+1,r7
aali_skip6:

	ldd r16,Y+2	;a
	ldd r17,Y+3
	ldd r18,Y+14*2	;b
	ldd r19,Y+14*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali7
	rjmp aali_skip7
aali7: bca_acs
	std Y+2,r6
	std Y+3,r7
	aca_bcs
	std Y+14*2,r6
	std Y+14*2+1,r7
aali_skip7:

	ld r16,Y	;a
	ldd r17,Y+1
	ldd r18,Y+15*2	;b
	ldd r19,Y+15*2+1
	lpm r20,Z+	;cs
	lpm r21,Z+
	lpm r22,Z+	;ca
	lpm r23,Z+

	mov r0,r16
	or r0,r17
	or r0,r18
	or r0,r19
	brne aali8
	rjmp aali_skip8
aali8: bca_acs
	st Y,r6
	std Y+1,r7
	aca_bcs
	std Y+15*2,r6
	std Y+15*2+1,r7
aali_skip8:

	dec r24
	breq aalix
	adiw Y,18*2
	rjmp aali_lp
aalix: ret

.equ a0=0
.equ a1=2
.equ a2=4
.equ a3=6
.equ a4=8
.equ a5=10
.equ a6=12
.equ a7=14
.equ a8=16
.equ a9=18
.equ a10=20
.equ a11=22
.equ a12=24
.equ a13=26
.equ a14=28
.equ a15=30
.equ a16=32
.equ a17=34
.equ a18=36
.equ a19=38
.equ a20=40
.equ a21=42
.equ a22=44
.equ a23=46
.equ a24=48
.equ a25=50

.equ m0=0
.equ m1=2
.equ m2=4
.equ m3=6
.equ m4=8
.equ m5=10
.equ m6=12
.equ m7=14

.equ x0=0
.equ x1=2
.equ x2=4
.equ x3=6
.equ x4=8
.equ x5=10
.equ x6=12
.equ x7=14
.equ x8=16

	.if fsdctmode==1

	.macro fsdctmul
	lpm r18,Z+
	lpm r19,Z+
	fmulsu r17,r19
	movw r21:r20,r1:r0
	fmul r16,r18
	adc r20,r15
	mov r3,r1
	fmul r16,r19
	adc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	fmulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	com r21
	neg r20
	sbci r21,-1
	st X+,r20
	st X+,r21
	.endmacro

	.else

	.macro fsdctmul
	rcall fsdctmul_
	.endmacro

fsdctmul_: lpm r18,Z+
	lpm r19,Z+
	fmulsu r17,r19
	movw r21:r20,r1:r0
	fmul r16,r18
	adc r20,r15
	mov r3,r1
	fmul r16,r19
	adc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	fmulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	com r21
	neg r20
	sbci r21,-1
	st X+,r20
	st X+,r21
	ret

	.endif

	.macro fsdctmul1
	lpm r18,Z+
	lpm r19,Z+
	fmulsu r17,r19
	movw r21:r20,r1:r0
	fmul r16,r18
	adc r20,r15
	mov r3,r1
	fmul r16,r19
	adc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	fmulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st X+,r20
	st X+,r21
	.endmacro

;reference implementation of fastsdct taken from libmad 0.15
;ZH:ZL - ^X
;XH:XL - ^Y
fastsdct:
	push XL
	push XH

	ldi YL,LOW(a)
	ldi YH,HIGH(a)
	push ZL
	push ZH

	ldd r2,Z+x3
	ldd r3,Z+x3+1
	ldd r4,Z+x5
	ldd r5,Z+x5+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a0,r0
	std Y+a0+1,r1
	sub r2,r4
	sbc r3,r5
	std Y+a1,r2
	std Y+a1+1,r3

	ldd r2,Z+x6
	ldd r3,Z+x6+1
	ldd r4,Z+x2
	ldd r5,Z+x2+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a2,r0
	std Y+a2+1,r1
	sub r2,r4
	sbc r3,r5
	std Y+a3,r2
	std Y+a3+1,r3

	ldd r2,Z+x1
	ldd r3,Z+x1+1
	ldd r4,Z+x7
	ldd r5,Z+x7+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a4,r0
	std Y+a4+1,r1
	sub r2,r4
	sbc r3,r5
	std Y+a5,r2
	std Y+a5+1,r3

	ldd r2,Z+x8
	ldd r3,Z+x8+1
	ldd r4,Z+x0
	ldd r5,Z+x0+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a6,r0
	std Y+a6+1,r1
	sub r2,r4
	sbc r3,r5
	std Y+a7,r2
	std Y+a7+1,r3

	ldd r2,Y+a0
	ldd r3,Y+a0+1
	ldd r4,Y+a2
	ldd r5,Y+a2+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a8,r0
	std Y+a8+1,r1
	movw r1:r0,r3:r2
	sub r0,r4
	sbc r1,r5
	std Y+a9,r0
	std Y+a9+1,r1
	ldd r6,Y+a6
	ldd r7,Y+a6+1
	sub r2,r6
	sbc r3,r7
	std Y+a10,r2
	std Y+a10+1,r3
	sub r4,r6
	sbc r5,r7
	std Y+a11,r4
	std Y+a11+1,r5
	ldd r2,Y+a8
	ldd r3,Y+a8+1
	add r2,r6
	adc r3,r7
	std Y+a12,r2
	std Y+a12+1,r3
	ldd r2,Y+a1
	ldd r3,Y+a1+1
	ldd r4,Y+a3
	ldd r5,Y+a3+1
	movw r1:r0,r3:r2
	sub r0,r4
	sbc r1,r5
	std Y+a13,r0
	std Y+a13+1,r1
	ldd r6,Y+a7
	ldd r7,Y+a7+1
	add r0,r6
	adc r1,r7
	std Y+a14,r0
	std Y+a14+1,r1
	movw r1:r0,r5:r4
	add r0,r6
	adc r1,r7
	std Y+a15,r0
	std Y+a15+1,r1
	movw r1:r0,r3:r2
	sub r0,r6
	sbc r1,r7
	std Y+a16,r0
	std Y+a16+1,r1
	add r2,r4
	adc r3,r5
	std Y+a17,r2
	std Y+a17+1,r3

	ldi ZL,LOW(sdctconst*2)
	ldi ZH,HIGH(sdctconst*2)
	ldi XL,LOW(m)
	ldi XH,HIGH(m)

	ldd r16,Y+a17
	ldd r17,Y+a17+1
	fsdctmul

	ldd r16,Y+a16
	ldd r17,Y+a16+1
	fsdctmul

	ldd r16,Y+a15
	ldd r17,Y+a15+1
	fsdctmul

	ldd r16,Y+a14
	ldd r17,Y+a14+1
	fsdctmul

	ldd r16,Y+a5
	ldd r17,Y+a5+1
	fsdctmul

	ldd r16,Y+a11
	ldd r17,Y+a11+1
	fsdctmul1

	ldd r16,Y+a10
	ldd r17,Y+a10+1
	fsdctmul

	ldd r16,Y+a9
	ldd r17,Y+a9+1
	fsdctmul

	pop ZH
	pop ZL
	ldd r2,Z+x4
	ldd r3,Z+x4+1
	ldd r4,Y+a4
	ldd r5,Y+a4+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a18,r0
	std Y+a18+1,r1
	add r2,r2
	adc r3,r3
	sub r2,r4
	sbc r3,r5
	std Y+a19,r2
	std Y+a19+1,r3
	ldi ZL,LOW(m)
	ldi ZH,HIGH(m)
	ldd r4,Z+m5
	ldd r5,Z+m5+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a20,r0
	std Y+a20+1,r1
	movw r1:r0,r3:r2
	sub r0,r4
	sbc r1,r5
	std Y+a21,r0
	std Y+a21+1,r1
	ldd r4,Z+m6
	ldd r5,Z+m6+1
	add r2,r4
	adc r3,r5
	std Y+a22,r2
	std Y+a22+1,r3
	ldd r2,Z+m4
	ldd r3,Z+m4+1
	ldd r4,Z+m2
	ldd r5,Z+m2+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	std Y+a23,r0
	std Y+a23+1,r1
	movw r1:r0,r3:r2
	sub r0,r4
	sbc r1,r5
	std Y+a24,r0
	std Y+a24+1,r1
	ldd r4,Z+m1
	ldd r5,Z+m1+1
	add r2,r4
	adc r3,r5
	std Y+a25,r2
	std Y+a25+1,r3

	pop XH
	pop XL

	ldd r2,Y+a18
	ldd r3,Y+a18+1
	ldd r4,Y+a12
	ldd r5,Y+a12+1
	movw r1:r0,r3:r2
	add r0,r4
	adc r1,r5
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r6,Z+m0
	ldd r7,Z+m0+1
	ldd r8,Y+a25
	ldd r9,Y+a25+1
	movw r1:r0,r7:r6
	sub r0,r8
	sbc r1,r9
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r8,Z+m7
	ldd r9,Z+m7+1
	ldd r10,Y+a20
	ldd r11,Y+a20+1
	movw r1:r0,r9:r8
	sub r0,r10
	sbc r1,r11
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r0,Z+m3
	ldd r1,Z+m3+1
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r0,Y+a21
	ldd r1,Y+a21+1
	ldd r12,Z+m6
	ldd r13,Z+m6+1
	sub r0,r12
	sbc r1,r13
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r0,Y+a24
	ldd r1,Y+a24+1
	ldd r12,Z+m1
	ldd r13,Z+m1+1
	sub r0,r12
	sbc r1,r13
	st X+,r0
	st X+,r1
	adiw X,2
	add r2,r2
	adc r3,r3
	sub r4,r2
	sbc r5,r3
	st X+,r4
	st X+,r5
	adiw X,2
	ldd r0,Y+a23
	ldd r1,Y+a23+1
	add r0,r6
	adc r1,r7
	st X+,r0
	st X+,r1
	adiw X,2
	ldd r0,Y+a22
	ldd r1,Y+a22+1
	add r0,r8
	adc r1,r9
	st X+,r0
	st X+,r1
	ret

	.if sdct2mode==1

	.macro sdctmul
	sub r16,r0
	sbc r17,r1
	lpm r18,Z+
	lpm r19,Z+
	fmulsu r17,r19
	movw r21:r20,r1:r0
	fmul r16,r18
	adc r20,r15
	mov r3,r1
	fmul r16,r19
	adc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	fmulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st X+,r20
	st X+,r21
	.endmacro

	.else

	.macro sdctmul
	rcall sdctmul_
	.endmacro

sdctmul_: sub r16,r0
	sbc r17,r1
	lpm r18,Z+
	lpm r19,Z+
	fmulsu r17,r19
	movw r21:r20,r1:r0
	fmul r16,r18
	adc r20,r15
	mov r3,r1
	fmul r16,r19
	adc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	fmulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st X+,r20
	st X+,r21
	ret

	.endif

;reference implementation of sdct2 taken from libmad 0.15
;YH:YL - ^x
;r25:r24 - ^X
sdct2:
	push YL
	push YH
	ldi XL,LOW(tmp)
	ldi XH,HIGH(tmp)
	ldd r0,Y+0
	ldd r1,Y+1
	ldd r2,Y+17*2
	ldd r3,Y+17*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+2
	ldd r1,Y+3
	ldd r2,Y+16*2
	ldd r3,Y+16*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+4
	ldd r1,Y+5
	ldd r2,Y+15*2
	ldd r3,Y+15*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+6
	ldd r1,Y+7
	ldd r2,Y+14*2
	ldd r3,Y+14*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+8
	ldd r1,Y+9
	ldd r2,Y+13*2
	ldd r3,Y+13*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+10
	ldd r1,Y+11
	ldd r2,Y+12*2
	ldd r3,Y+12*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+12
	ldd r1,Y+13
	ldd r2,Y+11*2
	ldd r3,Y+11*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+14
	ldd r1,Y+15
	ldd r2,Y+10*2
	ldd r3,Y+10*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1
	ldd r0,Y+16
	ldd r1,Y+17
	ldd r2,Y+9*2
	ldd r3,Y+9*2+1
	add r0,r2
	adc r1,r3
	st X+,r0
	st X+,r1

	ldi ZL,LOW(tmp)
	ldi ZH,HIGH(tmp)
	movw X,r25:r24
	rcall fastsdct

	pop YH
	pop YL
	ldi ZL,LOW(sdctscale*2)
	ldi ZH,HIGH(sdctscale*2)
	ldi XL,LOW(tmp)
	ldi XH,HIGH(tmp)
	ldd r16,Y+0
	ldd r17,Y+1
	ldd r0,Y+17*2
	ldd r1,Y+17*2+1
	sdctmul
	ldd r16,Y+2
	ldd r17,Y+3
	ldd r0,Y+16*2
	ldd r1,Y+16*2+1
	sdctmul
	ldd r16,Y+4
	ldd r17,Y+5
	ldd r0,Y+15*2
	ldd r1,Y+15*2+1
	sdctmul
	ldd r16,Y+6
	ldd r17,Y+7
	ldd r0,Y+14*2
	ldd r1,Y+14*2+1
	sdctmul
	ldd r16,Y+8
	ldd r17,Y+9
	ldd r0,Y+13*2
	ldd r1,Y+13*2+1
	sdctmul
	ldd r16,Y+10
	ldd r17,Y+11
	ldd r0,Y+12*2
	ldd r1,Y+12*2+1
	sdctmul
	ldd r16,Y+12
	ldd r17,Y+13
	ldd r0,Y+11*2
	ldd r1,Y+11*2+1
	sdctmul
	ldd r16,Y+14
	ldd r17,Y+15
	ldd r0,Y+10*2
	ldd r1,Y+10*2+1
	sdctmul
	ldd r16,Y+16
	ldd r17,Y+17
	ldd r0,Y+9*2
	ldd r1,Y+9*2+1
	sdctmul

	ldi ZL,LOW(tmp)
	ldi ZH,HIGH(tmp)
	movw X,r25:r24
	adiw X,2
	rcall fastsdct

	movw Y,r25:r24
	ldd r0,Y+3*2
	ldd r1,Y+3*2+1
	ldd r2,Y+1*2
	ldd r3,Y+1*2+1
	sub r0,r2
	sbc r1,r3
	std Y+3*2,r0
	std Y+3*2+1,r1
	ldd r0,Y+5*2
	ldd r1,Y+5*2+1
	ldd r2,Y+3*2
	ldd r3,Y+3*2+1
	sub r0,r2
	sbc r1,r3
	std Y+5*2,r0
	std Y+5*2+1,r1
	ldd r0,Y+7*2
	ldd r1,Y+7*2+1
	ldd r2,Y+5*2
	ldd r3,Y+5*2+1
	sub r0,r2
	sbc r1,r3
	std Y+7*2,r0
	std Y+7*2+1,r1
	ldd r0,Y+9*2
	ldd r1,Y+9*2+1
	ldd r2,Y+7*2
	ldd r3,Y+7*2+1
	sub r0,r2
	sbc r1,r3
	std Y+9*2,r0
	std Y+9*2+1,r1
	ldd r0,Y+11*2
	ldd r1,Y+11*2+1
	ldd r2,Y+9*2
	ldd r3,Y+9*2+1
	sub r0,r2
	sbc r1,r3
	std Y+11*2,r0
	std Y+11*2+1,r1
	ldd r0,Y+13*2
	ldd r1,Y+13*2+1
	ldd r2,Y+11*2
	ldd r3,Y+11*2+1
	sub r0,r2
	sbc r1,r3
	std Y+13*2,r0
	std Y+13*2+1,r1
	ldd r0,Y+15*2
	ldd r1,Y+15*2+1
	ldd r2,Y+13*2
	ldd r3,Y+13*2+1
	sub r0,r2
	sbc r1,r3
	std Y+15*2,r0
	std Y+15*2+1,r1
	ldd r0,Y+17*2
	ldd r1,Y+17*2+1
	ldd r2,Y+15*2
	ldd r3,Y+15*2+1
	sub r0,r2
	sbc r1,r3
	std Y+17*2,r0
	std Y+17*2+1,r1
	ret

	.if dctmode==1

	.macro dctmul
	ld r16,X+
	ld r17,X+
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st Y+,r20
	st Y+,r21
	.endmacro

	.else

	.macro dctmul
	rcall dctmul_
	.endmacro

dctmul_: ld r16,X+
	ld r17,X+
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st Y+,r20
	st Y+,r21
	ret

	.endif

;reference implementation of dct4 taken from libmad 0.15
;XH:XL - ^y
dct4:
	ldi ZL,LOW(dctscale*2)
	ldi ZH,HIGH(dctscale*2)
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	dctmul
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	ldi r24,LOW(tmp2)
	ldi r25,HIGH(tmp2)
	rcall sdct2

	ldi YL,LOW(tmp2)
	ldi YH,HIGH(tmp2)
	ldd r0,Y+0
	ldd r1,Y+1
	asr r1
	ror r0
	std Y+0,r0
	std Y+1,r1
	ldd r2,Y+1*2
	ldd r3,Y+1*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+1*2,r2
	std Y+1*2+1,r3
	ldd r0,Y+2*2
	ldd r1,Y+2*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+2*2,r0
	std Y+2*2+1,r1
	ldd r2,Y+3*2
	ldd r3,Y+3*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+3*2,r2
	std Y+3*2+1,r3
	ldd r0,Y+4*2
	ldd r1,Y+4*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+4*2,r0
	std Y+4*2+1,r1
	ldd r2,Y+5*2
	ldd r3,Y+5*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+5*2,r2
	std Y+5*2+1,r3
	ldd r0,Y+6*2
	ldd r1,Y+6*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+6*2,r0
	std Y+6*2+1,r1
	ldd r2,Y+7*2
	ldd r3,Y+7*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+7*2,r2
	std Y+7*2+1,r3
	ldd r0,Y+8*2
	ldd r1,Y+8*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+8*2,r0
	std Y+8*2+1,r1
	ldd r2,Y+9*2
	ldd r3,Y+9*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+9*2,r2
	std Y+9*2+1,r3
	ldd r0,Y+10*2
	ldd r1,Y+10*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+10*2,r0
	std Y+10*2+1,r1
	ldd r2,Y+11*2
	ldd r3,Y+11*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+11*2,r2
	std Y+11*2+1,r3
	ldd r0,Y+12*2
	ldd r1,Y+12*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+12*2,r0
	std Y+12*2+1,r1
	ldd r2,Y+13*2
	ldd r3,Y+13*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+13*2,r2
	std Y+13*2+1,r3
	ldd r0,Y+14*2
	ldd r1,Y+14*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+14*2,r0
	std Y+14*2+1,r1
	ldd r2,Y+15*2
	ldd r3,Y+15*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+15*2,r2
	std Y+15*2+1,r3
	ldd r0,Y+16*2
	ldd r1,Y+16*2+1
	asr r1
	ror r0
	sub r0,r2
	sbc r1,r3
	std Y+16*2,r0
	std Y+16*2+1,r1
	ldd r2,Y+17*2
	ldd r3,Y+17*2+1
	asr r3
	ror r2
	sub r2,r0
	sbc r3,r1
	std Y+17*2,r2
	std Y+17*2+1,r3
	ret

;reference implementation of imdct36 taken from libmad 0.15, it uses Szu-Wei Lee's fast algorithm
;XH:XL - ^x
imdct36:
	rcall dct4
	ldi YL,LOW(_z)
	ldi YH,HIGH(_z)
	ldi ZL,LOW(tmp2)
	ldi ZH,HIGH(tmp2)

	ldd r0,Z+18+0
	ldd r1,Z+18+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+1*2
	ldd r1,Z+18+1*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+2*2
	ldd r1,Z+18+2*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+3*2
	ldd r1,Z+18+3*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+4*2
	ldd r1,Z+18+4*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+5*2
	ldd r1,Z+18+5*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+6*2
	ldd r1,Z+18+6*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+7*2
	ldd r1,Z+18+7*2+1
	st Y+,r0
	st Y+,r1
	ldd r0,Z+18+8*2
	ldd r1,Z+18+8*2+1
	st Y+,r0
	st Y+,r1

	ldd r16,Z+17*2
	ldd r17,Z+17*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+16*2
	ldd r17,Z+16*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+15*2
	ldd r17,Z+15*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+14*2
	ldd r17,Z+14*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+13*2
	ldd r17,Z+13*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+12*2
	ldd r17,Z+12*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+11*2
	ldd r17,Z+11*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+10*2
	ldd r17,Z+10*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+9*2
	ldd r17,Z+9*2+1
	com r17
	neg r16
	sbci r17,-1
	st Y+,r16
	st Y+,r17
	ldd r16,Z+8*2
	ldd r17,Z+8*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+0,r16
	std Y+1,r17
	std Y+18+8*2,r16
	std Y+18+8*2+1,r17
	ldd r16,Z+7*2
	ldd r17,Z+7*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+1*2,r16
	std Y+1*2+1,r17
	std Y+18+7*2,r16
	std Y+18+7*2+1,r17
	ldd r16,Z+6*2
	ldd r17,Z+6*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+2*2,r16
	std Y+2*2+1,r17
	std Y+18+6*2,r16
	std Y+18+6*2+1,r17
	ldd r16,Z+5*2
	ldd r17,Z+5*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+3*2,r16
	std Y+3*2+1,r17
	std Y+18+5*2,r16
	std Y+18+5*2+1,r17
	ldd r16,Z+4*2
	ldd r17,Z+4*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+4*2,r16
	std Y+4*2+1,r17
	std Y+18+4*2,r16
	std Y+18+4*2+1,r17
	ldd r16,Z+3*2
	ldd r17,Z+3*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+5*2,r16
	std Y+5*2+1,r17
	std Y+18+3*2,r16
	std Y+18+3*2+1,r17
	ldd r16,Z+2*2
	ldd r17,Z+2*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+6*2,r16
	std Y+6*2+1,r17
	std Y+18+2*2,r16
	std Y+18+2*2+1,r17
	ldd r16,Z+1*2
	ldd r17,Z+1*2+1
	com r17
	neg r16
	sbci r17,-1
	std Y+7*2,r16
	std Y+7*2+1,r17
	std Y+18+1*2,r16
	std Y+18+1*2+1,r17
	ldd r16,Z+0
	ldd r17,Z+1
	com r17
	neg r16
	sbci r17,-1
	std Y+8*2,r16
	std Y+8*2+1,r17
	std Y+18,r16
	std Y+18+1,r17
	ret


	.if eval_hd==1

	.if imdctmode==1

	.macro idctsmul
	ld r16,X+
	ld r17,X+
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	muls r17,r19
	movw r21:r20,r1:r0
	mulsu r19,r16
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	.endmacro

	.macro idctsmulo
	mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	.endmacro

	.else

	.macro idctsmul
	rcall idctsmul_
	.endmacro
	.macro idctsmulo
	rcall idctsmulo_
	.endmacro

idctsmul_:	ld r16,X+
	ld r17,X+
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	muls r17,r19
	movw r21:r20,r1:r0
	mulsu r19,r16
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	ret

idctsmulo_: mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	ret

	.endif

;reference implementation of imdct_s taken from libmad 0.15
;performs IMDCT and windowing for short blocks
;XH:XL - ^x
imdct_s:
	push ZL
	push ZH
	push YL
	push YH
	push XL
	push XH

	clr r23
imdct_s_lp: movw r9:r8,X
	ldi ZL,LOW(imdct_s_tbl*2)
	ldi ZH,HIGH(imdct_s_tbl*2)
	clr r22
imdct_s_lp1: movw X,r9:r8
	idctsmul
	movw r5:r4,r21:r20
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r20,r4
	adc r21,r5
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	mov r16,r22
	add r16,r23
	add YL,r16
	adc YH,r15
	st Y+,r20
	st Y,r21
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	ldi r16,5*2
	sub r16,r22
	add r16,r23
	add YL,r16
	adc YH,r15
	com r21
	neg r20
	sbci r21,-1
	st Y+,r20
	st Y,r21

	movw X,r9:r8
	idctsmul
	movw r5:r4,r21:r20
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r4,r20
	adc r5,r21
	idctsmul
	add r20,r4
	adc r21,r5
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	ldi r16,6*2
	add r16,r22
	add r16,r23
	add YL,r16
	adc YH,r15
	st Y+,r20
	st Y,r21
	ldi YL,LOW(tmp1)
	ldi YH,HIGH(tmp1)
	ldi r16,11*2
	sub r16,r22
	add r16,r23
	add YL,r16
	adc YH,r15
	st Y+,r20
	st Y,r21
	subi r22,-2
	cpi r22,6
	breq imdct_s1
	rjmp imdct_s_lp1
imdct_s1: subi r23,-12*2
	cpi r23,36*2
	breq imdct_s2
	rjmp imdct_s_lp
imdct_s2:
	ldi XL,LOW(tmp1)
	ldi XH,HIGH(tmp1)
	ldi YL,LOW(_z)
	ldi YH,HIGH(_z)
	ldi ZL,LOW(window_s*2)
	ldi ZH,HIGH(window_s*2)
	ldi r22,6
imdct_ovw_lp:
	lpm r4,Z+				;wptr[0]
	lpm r5,Z+
	adiw Z,5*2
	lpm r6,Z+				;wptr[6]
	lpm r7,Z+
	sbiw Z,6*2

	st Y+,r15				;z[i+0]=0
	st Y+,r15

	ld r16,X+				;yptr[0+0]
	ld r17,X+
	movw r19:r18,r5:r4
	idctsmulo
	std Y+5*2,r20			;z[i+6]
	std Y+5*2+1,r21

	movw r11:r10,Z
	movw Z,X
	ldd r16,Z+5*2			;yptr[0+6]
	ldd r17,Z+5*2+1
	movw r19:r18,r7:r6
	idctsmulo
	movw r9:r8,r21:r20
	ldd r16,Z+11*2			;yptr[12+0]
	ldd r17,Z+11*2+1
	movw r19:r18,r5:r4
	idctsmulo
	add r20,r8
	adc r21,r9
	std Y+11*2,r20			;z[i+12]
	std Y+11*2+1,r21

	ldd r16,Z+17*2			;yptr[12+6]
	ldd r17,Z+17*2+1
	movw r19:r18,r7:r6
	idctsmulo
	movw r9:r8,r21:r20
	ldd r16,Z+23*2			;yptr[24+0]
	ldd r17,Z+23*2+1
	movw r19:r18,r5:r4
	idctsmulo
	add r20,r8
	adc r21,r9
	std Y+17*2,r20			;z[i+18]
	std Y+17*2+1,r21

	ldd r16,Z+29*2			;yptr[24+6]
	ldd r17,Z+29*2+1
	movw r19:r18,r7:r6
	idctsmulo
	std Y+23*2,r20			;z[i+24]
	std Y+23*2+1,r21

	std Y+29*2,r15			;z[i+30]=0
	std Y+29*2+1,r15

	movw Z,r11:r10
	dec r22
	breq imdct_sx
	rjmp imdct_ovw_lp
imdct_sx:

	pop XH
	pop XL
	adiw X,18*2
	pop YH
	pop YL
	pop ZH
	pop ZL
	ret

	.endif

	.if ovlmode==1

	.macro ovl1
	ld r0,Y+
	ld r1,Y+
	ld r2,X+
	ld r3,X+
	add r0,r2
	adc r1,r3
	add r0,r0
	adc r1,r1
	st Z+,r0
	st Z+,r1
	adiw Z,62
	.endmacro

	.macro ovl2
	ld r0,X+
	ld r1,X+
	st Y+,r0
	st Y+,r1
	.endmacro

	.else

	.macro ovl1
	rcall ovl1_
	.endmacro

ovl1_: ld r0,Y+
	ld r1,Y+
	ld r2,X+
	ld r3,X+
	add r0,r2
	adc r1,r3
	add r0,r0
	adc r1,r1
	st Z+,r0
	st Z+,r1
	adiw Z,62
	ret

	.macro ovl2
	rcall ovl2_
	.endmacro

ovl2_: ld r0,X+
	ld r1,X+
	st Y+,r0
	st Y+,r1
	ret

	.endif

;reference implementation of overlap function taken from libmad 0.15
overlap: push XL
	push XH
	push ZL
	push ZH
	push YL
	push YH
	ldi XL,LOW(_z)
	ldi XH,HIGH(_z)

	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1
	ovl1

	pop YH
	pop YL

	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	ovl2
	pop ZH
	pop ZL
	pop XH
	pop XL
	ret

;overlapped adding and odd-line subband frequency inversion for the whole long block output
;based on libmad 0.15
transform:
	ldi XL,LOW(frame)
	ldi XH,HIGH(frame)
	ldi YL,LOW(ovlb)
	ldi YH,HIGH(ovlb)
	ldi ZL,LOW(sample)
	ldi ZH,HIGH(sample)

	ldi r16,32
transform_lp: push r16
	.if eval_hd==1
	sbis btype,bt2			;skip if Block_type 2
	.endif
	rcall imdct_l
	.if eval_hd==1
	sbic btype,bt2			;skip if not Block_type 2
	rcall imdct_s
	.endif
	rcall overlap
	pop r16
	sbrc r16,0
	rcall freqinv
	adiw Z,2
	dec r16
	brne transform_lp
	ret

freqinv:
	push ZL
	push ZH
	subi ZL,-64
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	subi ZL,-128
	sbci ZH,-1
	ld r18,Z
	ldd r19,Z+1
	com r19
	neg r18
	sbci r19,-1
	st Z,r18
	std Z+1,r19
	pop ZH
	pop ZL
	ret

	.if imdctmode==1

	.macro idctmul
	ld r16,Y
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st Y+,r20
	st Y+,r21
	.endmacro

	.else

	.macro idctmul
	rcall idctmul_
	.endmacro

idctmul_: ld r16,Y
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	mul r16,r18
	mov r3,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r3,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r3,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r3,r3
	adc r20,r15
	adc r21,r15
	.endif
	st Y+,r20
	st Y+,r21
	ret

	.endif

;reference implementation of imdct_l taken from libmad 0.15
;performs IMDCT and windowing for long blocks
;XH:XL - ^x
imdct_l:
	push ZL
	push ZH
	push YL
	push YH
	push XL
	push XH
	rcall imdct36

	ldi ZL,LOW(window_l*2)
	ldi ZH,HIGH(window_l*2)
	ldi YL,LOW(_z)
	ldi YH,HIGH(_z)
	.if eval_hd==1
	sbis btype,altwnd			;skip if alternate window
	rjmp imdct_l_c1
	sbic btype,bt3
	rjmp imdct_l_b3				;handle stop block
	.endif
imdct_l_c1:
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	.if eval_hd==1
	sbic btype,altwnd
	rjmp imdct_l_b1				;handle start block
	.endif
imdct_l_c2:
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
imdct_l_x: pop XH
	pop XL
	adiw X,18*2
	pop YH
	pop YL
	pop ZH
	pop ZL
	ret

	.if eval_hd==1
imdct_l_b1: adiw Y,6*2			;z[18] to z[23] remain unchanged
	ldi ZL,LOW((window_s+6)*2)
	ldi ZH,HIGH((window_s+6)*2)
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	rjmp imdct_l_x

imdct_l_b3:
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	st Y+,r15
	ldi ZL,LOW(window_s*2)
	ldi ZH,HIGH(window_s*2)
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	idctmul
	adiw Y,6*2			;z[12] to z[17] remain unchanged
	ldi ZL,LOW((window_l+18)*2)
	ldi ZH,HIGH((window_l+18)*2)
	rjmp imdct_l_c2
	.endif

.equ costab1=65457 ;0.998795456
.equ costab2=65220 ;0.995184727
.equ costab3=64827 ;0.989176510
.equ costab4=64277 ;0.980785280
.equ costab5=63572 ;0.970031253
.equ costab6=62714 ;0.956940336
.equ costab7=61705 ;0.941544065
.equ costab8=60547 ;0.923879533
.equ costab9=59244 ;0.903989293
.equ costab10=57798 ;0.881921264
.equ costab11=56212 ;0.857728610
.equ costab12=54491 ;0.831469612
.equ costab13=52639 ;0.803207531
.equ costab14=50660 ;0.773010453
.equ costab15=48559 ;0.740951125
.equ costab16=46341 ;0.707106781
.equ costab17=44011 ;0.671558955
.equ costab18=41576 ;0.634393284
.equ costab19=39040 ;0.595699304
.equ costab20=36410 ;0.555570233
.equ costab21=33692 ;0.514102744
.equ costab22=30893 ;0.471396737
.equ costab23=28020 ;0.427555093
.equ costab24=25080 ;0.382683432
.equ costab25=22078 ;0.336889853
.equ costab26=19024 ;0.290284677
.equ costab27=15924 ;0.242980180
.equ costab28=12785 ;0.195090322
.equ costab29=9616 ;0.146730474
.equ costab30=6424 ;0.098017140
.equ costab31=3216 ;0.049067674

.equ in0=0
.equ in1=2
.equ in2=4
.equ in3=6
.equ in4=8
.equ in5=10
.equ in6=12
.equ in7=14
.equ in8=16
.equ in9=18
.equ in10=20
.equ in11=22
.equ in12=24
.equ in13=26
.equ in14=28
.equ in15=30
.equ in16=32
.equ in17=34
.equ in18=36
.equ in19=38
.equ in20=40
.equ in21=42
.equ in22=44
.equ in23=46
.equ in24=48
.equ in25=50
.equ in26=52
.equ in27=54
.equ in28=56
.equ in29=58
.equ in30=60
.equ in31=62

.equ t32=0
.equ t33=2
.equ t34=4
.equ t35=6
.equ t36=8
.equ t37=10
.equ t38=12
.equ t39=14
.equ t40=16
.equ t41=18
.equ t42=20
.equ t43=22
.equ t44=24
.equ t45=26
.equ t46=28
.equ t47=30
.equ t48=32
.equ t49=34
.equ t50=36
.equ t51=38
.equ t52=40
.equ t53=42
.equ t54=44
.equ t55=46
.equ t56=48
.equ t57=50
.equ t58=52
.equ t59=54
.equ t60=56
.equ t61=58
.equ t62=60
.equ t63=62

.equ t64=0
.equ t65=2
.equ t66=4
.equ t115=6
.equ t116=8
.equ t118=10
.equ t119=12
.equ t121=14
.equ t122=16
.equ t67=18
.equ t125=20
.equ t126=22
.equ t93=24
.equ t128=26
.equ t129=28
.equ t98=30
.equ t68=32
.equ t132=34
.equ t133=36
.equ t104=38
.equ t82=40
.equ t136=42
.equ t137=44
.equ t110=46
.equ t87=48
.equ t77=50
.equ t99=52
.equ t164=54

.equ t69=t33
.equ t89=t34
.equ t70=t35
.equ t90=t36
.equ t71=t37
.equ t91=t38
.equ t72=t39
.equ t92=t40
.equ t73=t41
.equ t94=t42
.equ t74=t43
.equ t95=t44
.equ t75=t45
.equ t96=t46
.equ t76=t47
.equ t97=t48
.equ t78=t50
.equ t100=t51
.equ t79=t52
.equ t101=t53
.equ t80=t54
.equ t102=t55
.equ t81=t56
.equ t103=t57
.equ t83=t59
.equ t106=t60
.equ t84=t61
.equ t107=t62
.equ t85=t63
.equ t108=t64
.equ t86=t65
.equ t109=t66
.equ t144=t73
.equ t145=t74
.equ t146=t75
.equ t88=t76
.equ t148=t78
.equ t149=t79
.equ t150=t80
.equ t105=t81
.equ t152=t83
.equ t153=t84
.equ t154=t85
.equ t111=t86
.equ t157=t89
.equ t158=t90
.equ t159=t91
;.equ t127=t92
.equ t161=t94
.equ t162=t95
.equ t163=t96
.equ t112=t97
.equ t166=t100
.equ t167=t101
.equ t168=t102
.equ t171=t106
.equ t172=t107
.equ t173=t108
.equ t138=t109
.equ t139=t121
.equ t174=t136

	.if dct32mode==1

	.macro dctmul32
	mul r16,r18
	mov r2,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r2,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r2,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r2,r2
	adc r20,r15
	adc r21,r15
	.endif
	.endmacro

	.else

	.macro dctmul32
	rcall dctmul32_
	.endmacro

dctmul32_: mul r16,r18
	mov r2,r1
	mulsu r17,r19
	movw r21:r20,r1:r0
	mul r16,r19
	add r2,r0
	adc r20,r1
	adc r21,r15
	mulsu r17,r18
	sbc r21,r15
	add r2,r0
	adc r20,r1
	adc r21,r15
	.if round==1
	add r2,r2
	adc r20,r15
	adc r21,r15
	.endif
	ret

	.endif

;reference implementation of fast discrete cosine transform taken from libmad 0.15

;YH:YL - ^sample
;r25:r24 - ^lo+phase*32
dct32:
	ldi ZL,LOW(tm32)
	ldi ZH,HIGH(tm32)
	movw X,Z
	adiw X,t59

	ldd r2,Y+in0
	ldd r3,Y+in0+1
	ldd r4,Y+in31
	ldd r5,Y+in31+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t0
	adc r5,r3
	ldi r18,LOW(costab1)
	ldi r19,HIGH(costab1)
	dctmul32
	movw r7:r6,r21:r20	;t16
	ldd r2,Y+in15
	ldd r3,Y+in15+1
	ldd r8,Y+in16
	ldd r9,Y+in16+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t1
	adc r9,r3
	ldi r18,LOW(costab31)
	ldi r19,HIGH(costab31)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t41,r6
	std Z+t41+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab2)
	ldi r19,HIGH(costab2)
	dctmul32
	st X+,r20			;t59
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t33,r4
	std Z+t33+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t50,r20
	std Z+t50+1,r21

	ldd r2,Y+in7
	ldd r3,Y+in7+1
	ldd r4,Y+in24
	ldd r5,Y+in24+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t2
	adc r5,r3
	ldi r18,LOW(costab15)
	ldi r19,HIGH(costab15)
	dctmul32
	movw r7:r6,r21:r20	;t18
	ldd r2,Y+in8
	ldd r3,Y+in8+1
	ldd r8,Y+in23
	ldd r9,Y+in23+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t3
	adc r9,r3
	ldi r18,LOW(costab17)
	ldi r19,HIGH(costab17)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t42,r6
	std Z+t42+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab30)
	ldi r19,HIGH(costab30)
	dctmul32
	st X+,r20			;t60
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t34,r4
	std Z+t34+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t51,r20
	std Z+t51+1,r21

	ldd r2,Y+in3
	ldd r3,Y+in3+1
	ldd r4,Y+in28
	ldd r5,Y+in28+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t4
	adc r5,r3
	ldi r18,LOW(costab7)
	ldi r19,HIGH(costab7)
	dctmul32
	movw r7:r6,r21:r20	;t18
	ldd r2,Y+in12
	ldd r3,Y+in12+1
	ldd r8,Y+in19
	ldd r9,Y+in19+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t5
	adc r9,r3
	ldi r18,LOW(costab25)
	ldi r19,HIGH(costab25)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t43,r6
	std Z+t43+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab14)
	ldi r19,HIGH(costab14)
	dctmul32
	st X+,r20			;t61
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t35,r4
	std Z+t35+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t52,r20
	std Z+t52+1,r21

	ldd r2,Y+in4
	ldd r3,Y+in4+1
	ldd r4,Y+in27
	ldd r5,Y+in27+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t6
	adc r5,r3
	ldi r18,LOW(costab9)
	ldi r19,HIGH(costab9)
	dctmul32
	movw r7:r6,r21:r20	;t20
	ldd r2,Y+in11
	ldd r3,Y+in11+1
	ldd r8,Y+in20
	ldd r9,Y+in20+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t7
	adc r9,r3
	ldi r18,LOW(costab23)
	ldi r19,HIGH(costab23)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t44,r6
	std Z+t44+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab18)
	ldi r19,HIGH(costab18)
	dctmul32
	st X+,r20			;t62
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t36,r4
	std Z+t36+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t53,r20
	std Z+t53+1,r21

	ldd r2,Y+in1
	ldd r3,Y+in1+1
	ldd r4,Y+in30
	ldd r5,Y+in30+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t8
	adc r5,r3
	ldi r18,LOW(costab3)
	ldi r19,HIGH(costab3)
	dctmul32
	movw r7:r6,r21:r20	;t24
	ldd r2,Y+in14
	ldd r3,Y+in14+1
	ldd r8,Y+in17
	ldd r9,Y+in17+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t9
	adc r9,r3
	ldi r18,LOW(costab29)
	ldi r19,HIGH(costab29)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t45,r6
	std Z+t45+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab6)
	ldi r19,HIGH(costab6)
	dctmul32
	st X+,r20			;t63
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t37,r4
	std Z+t37+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t54,r20
	std Z+t54+1,r21

	ldd r2,Y+in6
	ldd r3,Y+in6+1
	ldd r4,Y+in25
	ldd r5,Y+in25+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t10
	adc r5,r3
	ldi r18,LOW(costab13)
	ldi r19,HIGH(costab13)
	dctmul32
	movw r7:r6,r21:r20	;t26
	ldd r2,Y+in9
	ldd r3,Y+in9+1
	ldd r8,Y+in22
	ldd r9,Y+in22+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t11
	adc r9,r3
	ldi r18,LOW(costab19)
	ldi r19,HIGH(costab19)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t46,r6
	std Z+t46+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab26)
	ldi r19,HIGH(costab26)
	dctmul32
	st X+,r20			;t64
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t38,r4
	std Z+t38+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t55,r20
	std Z+t55+1,r21

	ldd r2,Y+in2
	ldd r3,Y+in2+1
	ldd r4,Y+in29
	ldd r5,Y+in29+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t12
	adc r5,r3
	ldi r18,LOW(costab5)
	ldi r19,HIGH(costab5)
	dctmul32
	movw r7:r6,r21:r20	;t28
	ldd r2,Y+in13
	ldd r3,Y+in13+1
	ldd r8,Y+in18
	ldd r9,Y+in18+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t13
	adc r9,r3
	ldi r18,LOW(costab27)
	ldi r19,HIGH(costab27)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t47,r6
	std Z+t47+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab10)
	ldi r19,HIGH(costab10)
	dctmul32
	st X+,r20			;t65
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t39,r4
	std Z+t39+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t56,r20
	std Z+t56+1,r21

	ldd r2,Y+in5
	ldd r3,Y+in5+1
	ldd r4,Y+in26
	ldd r5,Y+in26+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r4,r2			;t14
	adc r5,r3
	ldi r18,LOW(costab11)
	ldi r19,HIGH(costab11)
	dctmul32
	movw r7:r6,r21:r20	;t30
	ldd r2,Y+in10
	ldd r3,Y+in10+1
	ldd r8,Y+in21
	ldd r9,Y+in21+1
	movw r17:r16,r3:r2
	sub r16,r8
	sbc r17,r9
	add r8,r2			;t15
	adc r9,r3
	ldi r18,LOW(costab21)
	ldi r19,HIGH(costab21)
	dctmul32
	movw r17:r16,r7:r6
	add r6,r20
	adc r7,r21
	std Z+t48,r6
	std Z+t48+1,r7
	sub r16,r20
	sbc r17,r21
	ldi r18,LOW(costab22)
	ldi r19,HIGH(costab22)
	dctmul32
	st X+,r20			;t66
	st X+,r21
	movw r17:r16,r5:r4
	add r4,r8
	adc r5,r9
	std Z+t40,r4
	std Z+t40+1,r5
	sub r16,r8
	sbc r17,r9
	dctmul32
	std Z+t57,r20
	std Z+t57+1,r21

	ldi YL,LOW(tm64)
	ldi YH,HIGH(tm64)

	ldi r18,LOW(costab4)
	ldi r19,HIGH(costab4)
	ldd r2,Z+t33
	ldd r3,Z+t33+1
	ldd r4,Z+t34
	ldd r5,Z+t34+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t69,r2
	std Z+t69+1,r3
	dctmul32
	std Z+t89,r20
	std Z+t89+1,r21

	ldd r2,Z+t41
	ldd r3,Z+t41+1
	ldd r4,Z+t42
	ldd r5,Z+t42+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t73,r2
	std Z+t73+1,r3
	dctmul32
	std Z+t94,r20
	std Z+t94+1,r21

	ldd r2,Z+t50
	ldd r3,Z+t50+1
	ldd r4,Z+t51
	ldd r5,Z+t51+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t78,r2
	std Z+t78+1,r3
	dctmul32
	std Z+t100,r20
	std Z+t100+1,r21

	ldd r2,Z+t59
	ldd r3,Z+t59+1
	ldd r4,Z+t60
	ldd r5,Z+t60+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t83,r2
	std Z+t83+1,r3
	dctmul32
	std Z+t106,r20
	std Z+t106+1,r21

	ldi r18,LOW(costab28)
	ldi r19,HIGH(costab28)
	ldd r2,Z+t35
	ldd r3,Z+t35+1
	ldd r4,Z+t36
	ldd r5,Z+t36+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t70,r2
	std Z+t70+1,r3
	dctmul32
	std Z+t90,r20
	std Z+t90+1,r21

	ldd r2,Z+t43
	ldd r3,Z+t43+1
	ldd r4,Z+t44
	ldd r5,Z+t44+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t74,r2
	std Z+t74+1,r3
	dctmul32
	std Z+t95,r20
	std Z+t95+1,r21

	ldd r2,Z+t52
	ldd r3,Z+t52+1
	ldd r4,Z+t53
	ldd r5,Z+t53+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t79,r2
	std Z+t79+1,r3
	dctmul32
	std Z+t101,r20
	std Z+t101+1,r21

	ldd r2,Z+t61
	ldd r3,Z+t61+1
	ldd r4,Z+t62
	ldd r5,Z+t62+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t84,r2
	std Z+t84+1,r3
	dctmul32
	std Z+t107,r20
	std Z+t107+1,r21

	ldi r18,LOW(costab12)
	ldi r19,HIGH(costab12)
	ldd r2,Z+t37
	ldd r3,Z+t37+1
	ldd r4,Z+t38
	ldd r5,Z+t38+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t71,r2
	std Z+t71+1,r3
	dctmul32
	std Z+t91,r20
	std Z+t91+1,r21

	ldd r2,Z+t45
	ldd r3,Z+t45+1
	ldd r4,Z+t46
	ldd r5,Z+t46+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t75,r2
	std Z+t75+1,r3
	dctmul32
	std Z+t96,r20
	std Z+t96+1,r21

	ldd r2,Z+t54
	ldd r3,Z+t54+1
	ldd r4,Z+t55
	ldd r5,Z+t55+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t80,r2
	std Z+t80+1,r3
	dctmul32
	std Z+t102,r20
	std Z+t102+1,r21

	ldd r2,Z+t63
	ldd r3,Z+t63+1
	ldd r4,Y+t64
	ldd r5,Y+t64+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t85,r2
	std Z+t85+1,r3
	dctmul32
	std Y+t108,r20
	std Y+t108+1,r21

	ldi r18,LOW(costab20)
	ldi r19,HIGH(costab20)
	ldd r2,Z+t39
	ldd r3,Z+t39+1
	ldd r4,Z+t40
	ldd r5,Z+t40+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t72,r2
	std Z+t72+1,r3
	dctmul32
	std Z+t92,r20
	std Z+t92+1,r21

	ldd r2,Z+t47
	ldd r3,Z+t47+1
	ldd r4,Z+t48
	ldd r5,Z+t48+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t76,r2
	std Z+t76+1,r3
	dctmul32
	std Z+t97,r20
	std Z+t97+1,r21

	ldd r2,Z+t56
	ldd r3,Z+t56+1
	ldd r4,Z+t57
	ldd r5,Z+t57+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Z+t81,r2
	std Z+t81+1,r3
	dctmul32
	std Z+t103,r20
	std Z+t103+1,r21

	ldd r2,Y+t65
	ldd r3,Y+t65+1
	ldd r4,Y+t66
	ldd r5,Y+t66+1
	movw r17:r16,r3:r2
	sub r16,r4
	sbc r17,r5
	add r2,r4
	adc r3,r5
	std Y+t86,r2
	std Y+t86+1,r3
	dctmul32
	std Y+t109,r20
	std Y+t109+1,r21

	ldd r2,Z+t69
	ldd r3,Z+t69+1
	ldd r4,Z+t70
	ldd r5,Z+t70+1
	add r4,r2		;t113
	adc r5,r3
	ldd r2,Z+t71
	ldd r3,Z+t71+1
	ldd r6,Z+t72
	ldd r7,Z+t72+1
	add r6,r2		;t114
	adc r7,r3

	movw r17:r16,r5:r4
	sub r16,r6
	sbc r17,r7
	add r4,r6
	adc r5,r7

	movw X,r25:r24
	subi XH,-2
	subi XL,-15*2		;hi[15]
	st X+,r4
	st X,r5

	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	movw X,r25:r24	;lo[0]
	st X+,r20
	st X,r21

	ldd r2,Z+t73
	ldd r3,Z+t73+1
	ldd r4,Z+t74
	ldd r5,Z+t74+1
	add r4,r2		;t115
	adc r5,r3
	std Y+t115,r4
	std Y+t115+1,r5
	ldd r2,Z+t75
	ldd r3,Z+t75+1
	ldd r6,Z+t76
	ldd r7,Z+t76+1
	add r6,r2		;t116
	adc r7,r3
	std Y+t116,r6
	std Y+t116+1,r7
	add r4,r6
	adc r5,r7
	std Z+t32,r4
	std Z+t32+1,r5
	movw r9:r8,r5:r4
	movw X,r25:r24
	subi XH,-2
	subi XL,-14*2		;hi[14]
	st X+,r4
	st X,r5

	ldd r2,Z+t78
	ldd r3,Z+t78+1
	ldd r4,Z+t79
	ldd r5,Z+t79+1
	add r4,r2		;t118
	adc r5,r3
	std Y+t118,r4
	std Y+t118+1,r5
	ldd r2,Z+t80
	ldd r3,Z+t80+1
	ldd r6,Z+t81
	ldd r7,Z+t81+1
	add r6,r2		;t119
	adc r7,r3
	std Y+t119,r6
	std Y+t119+1,r7
	add r4,r6
	adc r5,r7
	std Z+t58,r4
	std Z+t58+1,r5
	movw r11:r10,r5:r4
	movw X,r25:r24
	subi XH,-2
	subi XL,-13*2		;hi[13]
	st X+,r4
	st X,r5

	ldd r2,Z+t83
	ldd r3,Z+t83+1
	ldd r4,Z+t84
	ldd r5,Z+t84+1
	add r4,r2		;t121
	adc r5,r3
	std Y+t121,r4
	std Y+t121+1,r5
	ldd r2,Z+t85
	ldd r3,Z+t85+1
	ldd r6,Y+t86
	ldd r7,Y+t86+1
	add r6,r2		;t122
	adc r7,r3
	std Y+t122,r6
	std Y+t122+1,r7
	add r4,r6
	adc r5,r7
	std Y+t67,r4
	std Y+t67+1,r5
	movw r13:r12,r5:r4
	add r4,r4
	adc r5,r5
	sub r4,r8
	sbc r5,r9
	movw r9:r8,r5:r4
	std Z+t49,r4
	std Z+t49+1,r5
	movw X,r25:r24
	subi XH,-2
	subi XL,-12*2		;hi[12]
	st X+,r4
	st X,r5

	ldd r2,Z+t89
	ldd r3,Z+t89+1
	ldd r4,Z+t90
	ldd r5,Z+t90+1
	add r4,r2		;t125
	adc r5,r3
	std Y+t125,r4
	std Y+t125+1,r5
	ldd r2,Z+t91
	ldd r3,Z+t91+1
	ldd r6,Z+t92
	ldd r7,Z+t92+1
	add r6,r2		;t126
	adc r7,r3
	std Y+t126,r6
	std Y+t126+1,r7
	add r4,r6
	adc r5,r7
	std Y+t93,r4
	std Y+t93+1,r5
	movw X,r25:r24
	subi XH,-2
	subi XL,-11*2		;hi[11]
	st X+,r4
	st X,r5

	ldd r2,Z+t94
	ldd r3,Z+t94+1
	ldd r4,Z+t95
	ldd r5,Z+t95+1
	add r4,r2		;t128
	adc r5,r3
	std Y+t128,r4
	std Y+t128+1,r5
	ldd r2,Z+t96
	ldd r3,Z+t96+1
	ldd r6,Z+t97
	ldd r7,Z+t97+1
	add r6,r2		;t129
	adc r7,r3
	std Y+t129,r6
	std Y+t129+1,r7
	add r4,r6
	adc r5,r7
	std Y+t98,r4
	std Y+t98+1,r5
	add r4,r4
	adc r5,r5
	sub r4,r8
	sbc r5,r9
	std Y+t68,r4
	std Y+t68+1,r5
	movw r9:r8,r5:r4
	movw X,r25:r24
	subi XH,-2
	subi XL,-10*2		;hi[10]
	st X+,r4
	st X,r5

	ldd r2,Z+t100
	ldd r3,Z+t100+1
	ldd r4,Z+t101
	ldd r5,Z+t101+1
	add r4,r2		;t132
	adc r5,r3
	std Y+t132,r4
	std Y+t132+1,r5
	ldd r2,Z+t102
	ldd r3,Z+t102+1
	ldd r6,Z+t103
	ldd r7,Z+t103+1
	add r6,r2		;t133
	adc r7,r3
	std Y+t133,r6
	std Y+t133+1,r7
	add r4,r6
	adc r5,r7
	std Y+t104,r4
	std Y+t104+1,r5
	add r4,r4
	adc r5,r5
	sub r4,r10
	sbc r5,r11
	std Y+t82,r4
	std Y+t82+1,r5
	movw X,r25:r24
	subi XH,-2
	subi XL,-9*2		;hi[9]
	st X+,r4
	st X,r5

	ldd r2,Z+t106
	ldd r3,Z+t106+1
	ldd r4,Z+t107
	ldd r5,Z+t107+1
	add r4,r2		;t136
	adc r5,r3
	std Y+t136,r4
	std Y+t136+1,r5
	ldd r2,Y+t108
	ldd r3,Y+t108+1
	ldd r6,Y+t109
	ldd r7,Y+t109+1
	add r6,r2		;t137
	adc r7,r3
	std Y+t137,r6
	std Y+t137+1,r7
	add r4,r6
	adc r5,r7
	std Y+t110,r4
	std Y+t110+1,r5
	add r4,r4
	adc r5,r5
	sub r4,r12
	sbc r5,r13
	std Y+t87,r4
	std Y+t87+1,r5
	add r4,r4
	adc r5,r5
	sub r4,r8
	sbc r5,r9
	std Y+t77,r4
	std Y+t77+1,r5
	movw X,r25:r24
	subi XH,-2
	subi XL,-8*2		;hi[8]
	st X+,r4
	st X,r5

	ldd r16,Z+t69
	ldd r17,Z+t69+1
	ldd r2,Z+t70
	ldd r3,Z+t70+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t141
	ldd r16,Z+t71
	ldd r17,Z+t71+1
	ldd r2,Z+t72
	ldd r3,Z+t72+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	movw r7:r6,r21:r20	;t142
	add r20,r4
	adc r21,r5
	movw r9:r8,r21:r20	;t143
	movw X,r25:r24
	subi XH,-2
	subi XL,-7*2		;hi[7]
	st X+,r20
	st X,r21
	movw r17:r16,r5:r4
	sub r16,r6
	sbc r17,r7
	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r8
	sbc r21,r9
	movw X,r25:r24
	subi XL,-8*2		;lo[8]
	st X+,r20
	st X,r21

	ldd r16,Z+t73
	ldd r17,Z+t73+1
	ldd r2,Z+t74
	ldd r3,Z+t74+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t144
	std Z+t144,r20
	std Z+t144+1,r21
	ldd r16,Z+t75
	ldd r17,Z+t75+1
	ldd r2,Z+t76
	ldd r3,Z+t76+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t145,r20
	std Z+t145+1,r21
	add r20,r4
	adc r21,r5
	std Z+t146,r20
	std Z+t146+1,r21
	ldd r2,Y+t77
	ldd r3,Y+t77+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Z+t88,r20
	std Z+t88+1,r21
	movw X,r25:r24
	subi XH,-2
	subi XL,-6*2		;hi[6]
	st X+,r20
	st X,r21

	ldd r16,Z+t78
	ldd r17,Z+t78+1
	ldd r2,Z+t79
	ldd r3,Z+t79+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t148
	std Z+t148,r20
	std Z+t148+1,r21
	ldd r16,Z+t80
	ldd r17,Z+t80+1
	ldd r2,Z+t81
	ldd r3,Z+t81+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t149,r20
	std Z+t149+1,r21
	add r20,r4
	adc r21,r5
	std Z+t150,r20
	std Z+t150+1,r21
	ldd r2,Y+t82
	ldd r3,Y+t82+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Z+t105,r20
	std Z+t105+1,r21
	movw X,r25:r24
	subi XH,-2
	subi XL,-5*2		;hi[5]
	st X+,r20
	st X,r21

	ldd r16,Z+t83
	ldd r17,Z+t83+1
	ldd r2,Z+t84
	ldd r3,Z+t84+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t152
	std Z+t152,r20
	std Z+t152+1,r21
	ldd r16,Z+t85
	ldd r17,Z+t85+1
	ldd r2,Y+t86
	ldd r3,Y+t86+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t153,r20
	std Z+t153+1,r21
	add r20,r4
	adc r21,r5
	std Z+t154,r20
	std Z+t154+1,r21
	ldd r2,Y+t87
	ldd r3,Y+t87+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Y+t111,r20
	std Y+t111+1,r21
	ldd r2,Z+t88
	ldd r3,Z+t88+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Y+t99,r20
	std Y+t99+1,r21
	movw X,r25:r24
	subi XH,-2
	subi XL,-4*2		;hi[4]
	st X+,r20
	st X,r21

	ldd r16,Z+t89
	ldd r17,Z+t89+1
	ldd r2,Z+t90
	ldd r3,Z+t90+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t157
	std Z+t157,r20
	std Z+t157+1,r21
	ldd r16,Z+t91
	ldd r17,Z+t91+1
	ldd r2,Z+t92
	ldd r3,Z+t92+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t158,r20
	std Z+t158+1,r21
	add r20,r4
	adc r21,r5
	std Z+t159,r20
	std Z+t159+1,r21
	ldd r2,Y+t93
	ldd r3,Y+t93+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r7:r6,r21:r20	;t127
	movw X,r25:r24
	subi XH,-2
	subi XL,-3*2		;hi[3]
	st X+,r20
	st X,r21

	ldd r16,Y+t125
	ldd r17,Y+t125+1
	ldd r2,Y+t126
	ldd r3,Y+t126+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	movw r7:r6,r21:r20
	movw X,r25:r24
	subi XL,-4*2		;lo[4]
	st X+,r20
	st X,r21

	ldd r16,Z+t157
	ldd r17,Z+t157+1
	ldd r2,Z+t158
	ldd r3,Z+t158+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t159
	ldd r3,Z+t159+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	movw X,r25:r24
	subi XL,-12*2		;lo[12]
	st X+,r20
	st X,r21

	ldd r16,Z+t94
	ldd r17,Z+t94+1
	ldd r2,Z+t95
	ldd r3,Z+t95+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20	;t161
	std Z+t161,r20
	std Z+t161+1,r21
	ldd r16,Z+t96
	ldd r17,Z+t96+1
	ldd r2,Z+t97
	ldd r3,Z+t97+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t162,r20
	std Z+t162+1,r21
	add r20,r4
	adc r21,r5
	std Z+t163,r20
	std Z+t163+1,r21
	ldd r2,Y+t98
	ldd r3,Y+t98+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r7:r6,r21:r20	;t130
	ldd r2,Y+t99
	ldd r3,Y+t99+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Z+t112,r20
	std Z+t112+1,r21
	movw X,r25:r24
	subi XH,-2
	subi XL,-2*2		;hi[2]
	st X+,r20
	st X,r21
	
	ldd r16,Y+t128
	ldd r17,Y+t128+1
	ldd r2,Y+t129
	ldd r3,Y+t129+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	std Y+t164,r20
	std Y+t164+1,r21

	ldd r16,Z+t100
	ldd r17,Z+t100+1
	ldd r2,Z+t101
	ldd r3,Z+t101+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20
	std Z+t166,r20
	std Z+t166+1,r21
	ldd r16,Z+t102
	ldd r17,Z+t102+1
	ldd r2,Z+t103
	ldd r3,Z+t103+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t167,r20
	std Z+t167+1,r21
	add r20,r4
	adc r21,r5
	std Z+t168,r20
	std Z+t168+1,r21
	ldd r2,Y+t104
	ldd r3,Y+t104+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r7:r6,r21:r20	;t134
	ldd r2,Z+t105
	ldd r3,Z+t105+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r9:r8,r21:r20	;t120
	movw X,r25:r24
	subi XH,-2
	subi XL,-1*2		;hi[1]
	st X+,r20
	st X,r21

	ldd r16,Y+t118
	ldd r17,Y+t118+1
	ldd r2,Y+t119
	ldd r3,Y+t119+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r8
	sbc r21,r9
	movw r9:r8,r21:r20	;t135
	movw X,r25:r24
	subi XL,-2*2		;lo[2]
	st X+,r20
	st X,r21

	ldd r16,Y+t132
	ldd r17,Y+t132+1
	ldd r2,Y+t133
	ldd r3,Y+t133+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	movw r7:r6,r21:r20	;t169

	add r20,r20
	adc r21,r21
	sub r20,r8
	sbc r21,r9
	movw r9:r8,r21:r20	;t151
	movw X,r25:r24
	subi XL,-6*2		;lo[6]
	st X+,r20
	st X,r21	

	ldd r16,Z+t148
	ldd r17,Z+t148+1
	ldd r2,Z+t149
	ldd r3,Z+t149+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t150
	ldd r3,Z+t150+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r8
	sbc r21,r9
	movw r9:r8,r21:r20	;t170
	movw X,r25:r24
	subi XL,-10*2		;lo[10]
	st X+,r20
	st X,r21

	ldd r16,Z+t166
	ldd r17,Z+t166+1
	ldd r2,Z+t167
	ldd r3,Z+t167+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t168
	ldd r3,Z+t168+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	add r20,r20
	adc r21,r21
	sub r20,r8
	sbc r21,r9
	movw X,r25:r24
	subi XL,-14*2		;lo[14]
	st X+,r20
	st X,r21

	ldd r16,Z+t106
	ldd r17,Z+t106+1
	ldd r2,Z+t107
	ldd r3,Z+t107+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab8)
	ldi r19,HIGH(costab8)
	dctmul32
	movw r5:r4,r21:r20
	std Z+t171,r20
	std Z+t171+1,r21
	ldd r16,Y+t108
	ldd r17,Y+t108+1
	ldd r2,Y+t109
	ldd r3,Y+t109+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab24)
	ldi r19,HIGH(costab24)
	dctmul32
	std Z+t172,r20
	std Z+t172+1,r21
	add r20,r4
	adc r21,r5
	std Y+t173,r20
	std Y+t173+1,r21
	ldd r2,Y+t110
	ldd r3,Y+t110+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Y+t138,r20
	std Y+t138+1,r21
	ldd r2,Y+t111
	ldd r3,Y+t111+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r5:r4,r21:r20	;t123

	ldd r16,Y+t121
	ldd r17,Y+t121+1
	ldd r2,Y+t122
	ldd r3,Y+t122+1
	sub r16,r2
	sbc r17,r3
	ldi r18,LOW(costab16)
	ldi r19,HIGH(costab16)
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	std Y+t139,r20
	std Y+t139+1,r21
	
	ldd r2,Z+t112
	ldd r3,Z+t112+1
	add r4,r4
	adc r5,r5
	sub r4,r2		;t117
	sbc r5,r3
	movw X,r25:r24
	subi XH,-2
	st X+,r4		;hi[0]
	st X,r5

	ldd r16,Y+t115
	ldd r17,Y+t115+1
	ldd r2,Y+t116
	ldd r3,Y+t116+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	add r20,r20
	adc r21,r21
	sub r20,r4		;t124
	sbc r21,r5
	movw X,r25:r24
	subi XL,-1*2	;lo[1]
	st X+,r20
	st X,r21

	ldd r2,Y+t139
	ldd r3,Y+t139+1
	add r2,r2
	adc r3,r3
	sub r2,r20		;t131
	sbc r3,r21
	movw X,r25:r24
	subi XL,-3*2	;lo[3]
	st X+,r2
	st X,r3

	ldd r4,Y+t164
	ldd r5,Y+t164+1
	add r4,r4
	adc r5,r5
	sub r4,r2		;t140
	sbc r5,r3
	movw X,r25:r24
	subi XL,-5*2	;lo[5]
	st X+,r4
	st X,r5

	ldd r16,Y+t136
	ldd r17,Y+t136+1
	ldd r2,Y+t137
	ldd r3,Y+t137+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Y+t138
	ldd r3,Y+t138+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	std Y+t174,r20
	std Y+t174+1,r21
	ldd r2,Y+t139
	ldd r3,Y+t139+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	movw r7:r6,r21:r20	;t155
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	movw r5:r4,r21:r20	;t147
	movw X,r25:r24
	subi XL,-7*2	;lo[7]
	st X+,r20
	st X,r21
	
	ldd r16,Z+t144
	ldd r17,Z+t144+1
	ldd r2,Z+t145
	ldd r3,Z+t145+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t146
	ldd r3,Z+t146+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	movw r5:r4,r21:r20	;t156
	movw X,r25:r24
	subi XL,-9*2	;lo[9]
	st X+,r20
	st X,r21

	ldd r16,Z+t152
	ldd r17,Z+t152+1
	ldd r2,Z+t153
	ldd r3,Z+t153+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t154
	ldd r3,Z+t154+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	movw r7:r6,r21:r20	;t175
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	movw r5:r4,r21:r20	;t165
	movw X,r25:r24
	subi XL,-11*2	;lo[11]
	st X+,r20
	st X,r21

	ldd r16,Z+t161
	ldd r17,Z+t161+1
	ldd r2,Z+t162
	ldd r3,Z+t162+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Z+t163
	ldd r3,Z+t163+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	ldd r2,Y+t164
	ldd r3,Y+t164+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	movw r5:r4,r21:r20	;t176
	movw X,r25:r24
	subi XL,-13*2	;lo[13]
	st X+,r20
	st X,r21

	ldd r16,Z+t171
	ldd r17,Z+t171+1
	ldd r2,Z+t172
	ldd r3,Z+t172+1
	sub r16,r2
	sbc r17,r3
	dctmul32
	ldd r2,Y+t173
	ldd r3,Y+t173+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	ldd r2,Y+t174
	ldd r3,Y+t174+1
	add r20,r20
	adc r21,r21
	sub r20,r2
	sbc r21,r3
	add r20,r20
	adc r21,r21
	sub r20,r6
	sbc r21,r7
	add r20,r20
	adc r21,r21
	sub r20,r4
	sbc r21,r5
	movw X,r25:r24
	subi XL,-15*2	;lo[15]
	st X+,r20
	st X,r21
	ret

	.macro smul
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	muls r17,r19
	movw r5:r4,r1:r0
	mul r16,r18
	movw r3:r2,r1:r0
	mulsu r17,r18
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	mulsu r19,r16
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.endmacro

	.macro smac
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	muls r17,r19
	movw r9:r8,r1:r0
	mul r16,r18
	add r2,r0
	adc r3,r1
	adc r4,r8
	adc r5,r9
	mulsu r17,r18
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	mulsu r19,r16
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	.endmacro

	.macro sneg
	ldi r16,-1
	com r5
	com r4
	com r3
	neg r2
	sbc r3,r16
	sbc r4,r16
	sbc r5,r16
	.endmacro

	.macro shandle
	.if chip==1 || flash==1
	add r4,r4
	adc r5,r5
	add r4,r4
	adc r5,r5
	ldi r16,$80
	add r5,r16
	.else
	add r3,r3
	adc r4,r4
	adc r5,r5
	add r3,r3
	adc r4,r4
	adc r5,r5
	.endif
	.endmacro

	.macro sout
	shandle
	.if (chip!=1 && flash!=1) || use_dac==1
	st X+,r4
	.endif
	st X+,r5
	.endmacro

	.macro sout1
	shandle
	st -X,r5
	.if (chip!=1 && flash!=1) || use_dac==1
	st -X,r4
	.endif
	.endmacro

	.if sfmode==0
smul_: smul
	ret

smac_: 
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
smac_c: muls r17,r19
	movw r9:r8,r1:r0
	mul r16,r18
	add r2,r0
	adc r3,r1
	adc r4,r8
	adc r5,r9
	mulsu r17,r18
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	mulsu r19,r16
	sbc r5,r15
	add r3,r0
	adc r4,r1
	adc r5,r15
	ret

smac_id: 
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	add YL,r7
	adc YH,r15
	sbiw Z,3*2
	rjmp smac_c

smac_di:
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	sub YL,r7
	sbc YH,r15
	adiw Z,1*2
	rjmp smac_c

smac_ii:
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	add YL,r7
	adc YH,r15
	adiw Z,1*2
	rjmp smac_c

smac_dd:
	ldd r16,Y+0
	ldd r17,Y+1
	lpm r18,Z+
	lpm r19,Z+
	sub YL,r7
	sbc YH,r15
	sbiw Z,3*2
	rjmp smac_c

sneg_: sneg
	ret

sout_:
	movw X,r21:r20
	shandle
	.if (chip!=1 && flash!=1) || use_dac==1
	st X+,r4
	.endif
	st X+,r5
	movw r21:r20,X
	ret

	.endif

;reference implementation of polyphase filter taken from libmad 0.15
synthf: ldi r23,18				;number of samples
	ldi r20,LOW(output)
	ldi r21,HIGH(output)
	.if eval_hd==1
	sbis btype,cbuf
	ldi r20,LOW(output1)
	sbis btype,cbuf
	ldi r21,HIGH(output1)
	.endif
	ldi YL,LOW(sample)
	ldi YH,HIGH(sample)
	ldi r24,LOW(lo)
	ldi r25,HIGH(lo)
	lds r22,(phase)
	mov r16,r22
	swap r16
	add r24,r16
	add r24,r16
	adc r25,r15
synthf_lp: 
	push YL
	push YH
	push r20
	push r21
	call dct32
	pop r21
	pop r20
	push r20
	push r21
	push r24
	push r25
	ldi r16,64
	mov r7,r16
	mov r16,r22
	andi r16,14
	mov r12,r16
	add r12,r12		;pe
	mov r16,r22
	dec r16
	andi r16,15
	ori r16,1
	mov r13,r16
	add r13,r13		;po

	ldi ZL,LOW(synthcoeff*2)
	ldi ZH,HIGH(synthcoeff*2)
	add ZL,r13
	adc ZH,r15
	ldi YL,LOW(lo)
	ldi YH,HIGH(lo)
	sbrs r22,0
	subi YL,-32

	.if sfmode==1

	smul
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	smac
	add YL,r7		;sample 2
	adc YH,r15
	sbiw Z,3*2	;d+12
	smac
	add YL,r7		;sample 3
	adc YH,r15
	sbiw Z,3*2	;d+10
	smac
	add YL,r7		;sample 4
	adc YH,r15
	sbiw Z,3*2	;d+8
	smac
	add YL,r7		;sample 5
	adc YH,r15
	sbiw Z,3*2	;d+6
	smac
	add YL,r7		;sample 6
	adc YH,r15
	sbiw Z,3*2	;d+4
	smac
	add YL,r7		;sample 7
	adc YH,r15
	sbiw Z,3*2	;d+2
	smac
	sneg

	.else

	rcall smul_
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_
	rcall sneg_

	.endif

	ldi ZL,LOW(synthcoeff*2)
	ldi ZH,HIGH(synthcoeff*2)
	add ZL,r12
	adc ZH,r15
	ldi YL,LOW(lo)
	ldi YH,HIGH(lo)
	sbrc r22,0
	subi YL,-32

	.if sfmode==1
	smac
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	smac
	add YL,r7		;sample 2
	adc YH,r15
	sbiw Z,3*2	;d+12
	smac
	add YL,r7		;sample 3
	adc YH,r15
	sbiw Z,3*2	;d+10
	smac
	add YL,r7		;sample 4
	adc YH,r15
	sbiw Z,3*2	;d+8
	smac
	add YL,r7		;sample 5
	adc YH,r15
	sbiw Z,3*2	;d+6
	smac
	add YL,r7		;sample 6
	adc YH,r15
	sbiw Z,3*2	;d+4
	smac
	add YL,r7		;sample 7
	adc YH,r15
	sbiw Z,3*2	;d+2
	smac

	movw X,r21:r20
	sout
	movw r21:r20,X

	.else

	rcall smac_
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_
	rcall sout_

	.endif

	.if (chip==1 || flash==1) && use_dac!=1
	adiw X,31
	.else
	adiw X,31*2
	.endif
	movw r11:r10,X

	ldi r24,2

synthf_lp1:	
	ldi ZL,LOW(synthcoeff*2)
	ldi ZH,HIGH(synthcoeff*2)
	ldi r16,32
	mul r16,r24
	add r0,r13
	adc r1,r15
	add ZL,r0
	adc ZH,r1
	ldi YL,LOW(hi-2)
	ldi YH,HIGH(hi-2)
	add YL,r24
	adc YH,r15
	sbrs r22,0
	subi YL,-32

	.if sfmode==1

	smul
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	smac
	add YL,r7		;sample 2
	adc YH,r15
	sbiw Z,3*2	;d+12
	smac
	add YL,r7		;sample 3
	adc YH,r15
	sbiw Z,3*2	;d+10
	smac
	add YL,r7		;sample 4
	adc YH,r15
	sbiw Z,3*2	;d+8
	smac
	add YL,r7		;sample 5
	adc YH,r15
	sbiw Z,3*2	;d+6
	smac
	add YL,r7		;sample 6
	adc YH,r15
	sbiw Z,3*2	;d+4
	smac
	add YL,r7		;sample 7
	adc YH,r15
	sbiw Z,3*2	;d+2
	smac

	sneg

	.else

	rcall smul_
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_
	rcall sneg_

	.endif

	ldi ZL,LOW(synthcoeff*2+4)
	ldi ZH,HIGH(synthcoeff*2+4)
	ldi r16,32
	mul r16,r24
	add r0,r12
	adc r1,r15
	add ZL,r0
	adc ZH,r1
	ldi YL,LOW(lo+14*32)
	ldi YH,HIGH(lo+14*32)
	add YL,r24
	adc YH,r15
	sbrc r22,0
	subi YL,-32

	.if sfmode==1

	smac
	sub YL,r7		;sample 6
	sbc YH,r15
	adiw Z,1*2	;d+4
	smac
	sub YL,r7		;sample 5
	sbc YH,r15
	adiw Z,1*2	;d+6
	smac
	sub YL,r7		;sample 4
	sbc YH,r15
	adiw Z,1*2	;d+8
	smac
	sub YL,r7		;sample 3
	sbc YH,r15
	adiw Z,1*2	;d+10
	smac
	sub YL,r7		;sample 2
	sbc YH,r15
	adiw Z,1*2	;d+12
	smac
	sub YL,r7		;sample 1
	sbc YH,r15
	adiw Z,1*2	;d+14
	smac
	sub YL,r7		;sample 0
	sbc YH,r15
	sbiw Z,15*2	;d+0
	smac
	movw X,r21:r20
	sout
	movw r21:r20,X

	.else

	rcall smac_di
	rcall smac_di
	rcall smac_di
	rcall smac_di
	rcall smac_di
	rcall smac_di
	rcall smac_
	sub YL,r7		;sample 0
	sbc YH,r15
	sbiw Z,15*2	;d+0
	rcall smac_
	rcall sout_

	.endif

	ldi ZL,LOW(synthcoeff*2+30)
	ldi ZH,HIGH(synthcoeff*2+30)
	ldi r16,32
	mul r16,r24
	sub r0,r12
	sbc r1,r15
	add ZL,r0
	adc ZH,r1
	ldi YL,LOW(lo)
	ldi YH,HIGH(lo)
	add YL,r24
	adc YH,r15
	sbrc r22,0
	subi YL,-32

	.if sfmode==1

	smul
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,1*2	;d+17
	smac
	add YL,r7		;sample 2
	adc YH,r15
	adiw Z,1*2	;d+19
	smac
	add YL,r7		;sample 3
	adc YH,r15
	adiw Z,1*2	;d+21
	smac
	add YL,r7		;sample 4
	adc YH,r15
	adiw Z,1*2	;d+23
	smac
	add YL,r7		;sample 5
	adc YH,r15
	adiw Z,1*2	;d+25
	smac
	add YL,r7		;sample 6
	adc YH,r15
	adiw Z,1*2	;d+27
	smac
	add YL,r7		;sample 7
	adc YH,r15
	adiw Z,1*2	;d+29
	smac

	.else

	rcall smul_
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,1*2	;d+17
	rcall smac_ii
	rcall smac_ii
	rcall smac_ii
	rcall smac_ii
	rcall smac_ii
	rcall smac_ii
	rcall smac_

	.endif

	ldi ZL,LOW(synthcoeff*2+2*29)
	ldi ZH,HIGH(synthcoeff*2+2*29)
	ldi r16,32
	mul r16,r24
	sub r0,r13
	sbc r1,r15
	add ZL,r0
	adc ZH,r1
	ldi YL,LOW(hi-2+14*32)
	ldi YH,HIGH(hi-2+14*32)
	add YL,r24
	adc YH,r15
	sbrs r22,0
	subi YL,-32

	.if sfmode==1

	smac
	sub YL,r7		;sample 6
	sbc YH,r15
	sbiw Z,3*2	;d+27
	smac
	sub YL,r7		;sample 5
	sbc YH,r15
	sbiw Z,3*2	;d+25
	smac
	sub YL,r7		;sample 4
	sbc YH,r15
	sbiw Z,3*2	;d23
	smac
	sub YL,r7		;sample 3
	sbc YH,r15
	sbiw Z,3*2	;d+21
	smac
	sub YL,r7		;sample 2
	sbc YH,r15
	sbiw Z,3*2	;d+19
	smac
	sub YL,r7		;sample 1
	sbc YH,r15
	sbiw Z,3*2	;d+17
	smac
	sub YL,r7		;sample 0
	sbc YH,r15
	sbiw Z,3*2	;d+15
	smac

	.else

	rcall smac_dd
	rcall smac_dd
	rcall smac_dd
	rcall smac_dd
	rcall smac_dd
	rcall smac_dd
	rcall smac_dd
	rcall smac_

	.endif

	movw X,r11:r10
	sout1
	movw r11:r10,X

	subi r24,-2
	cpi r24,2*16
	breq synthf_x1
	rjmp synthf_lp1
synthf_x1:

	ldi ZL,LOW(synthcoeff*2+16*64)
	ldi ZH,HIGH(synthcoeff*2+16*64)
	add ZL,r13
	adc ZH,r15
	ldi YL,LOW(hi+2*15)
	ldi YH,HIGH(hi+2*15)
	sbrs r22,0
	subi YL,-32

	.if sfmode==1

	smul
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	smac
	add YL,r7		;sample 2
	adc YH,r15
	sbiw Z,3*2	;d+12
	smac
	add YL,r7		;sample 3
	adc YH,r15
	sbiw Z,3*2	;d+10
	smac
	add YL,r7		;sample 4
	adc YH,r15
	sbiw Z,3*2	;d+8
	smac
	add YL,r7		;sample 5
	adc YH,r15
	sbiw Z,3*2	;d+6
	smac
	add YL,r7		;sample 6
	adc YH,r15
	sbiw Z,3*2	;d+4
	smac
	add YL,r7		;sample 7
	adc YH,r15
	sbiw Z,3*2	;d+2
	smac
	sneg
	movw X,r21:r20
	sout

	.else

	rcall smul_
	add YL,r7		;sample 1
	adc YH,r15
	adiw Z,13*2	;d+14
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_id
	rcall smac_
	rcall sneg_
	rcall sout_

	.endif

	pop r25
	pop r24
	pop r21
	pop r20
	pop YH
	pop YL
	add YL,r7
	adc YH,r15
	.if (chip==1 || flash==1) && use_dac!=1
	subi r20,-32
	sbci r21,-1
	.else
	add r20,r7
	adc r21,r15
	.endif
	adiw r25:r24,32
	cpi r25,HIGH(hi)
	brne synthfa
	ldi r25,HIGH(lo)
synthfa: inc r22
	andi r22,15
	dec r23
	breq synthf_x
	jmp synthf_lp
synthf_x: sts (phase),r22
	ret

	.if log==1
dumpbyte: pop r0
	pop r1
	#if PROGMEM_SIZE>65535
	pop r2
	#endif
	pop r16
	#if PROGMEM_SIZE>65535
	push r2
	#endif
	push r1
	push r0
	push r16
	swap r16
	clr r0
	rcall dumpnibble
	pop r16
	inc r0
dumpnibble: andi r16,15
	cpi r16,10
	brcs dumpnibble1
	subi r16,-7
dumpnibble1: subi r16,-$30
	out PORTA,r0
	out PORTA,r16
	ret
	.endif

	.if eval_hd==1
	.if use_dac==1
sfrq_tbl:
	.dw int(clock*(1000/sfrq_mul)/11)
	.dw int(clock*(1000/sfrq_mul)/12)
	.dw int(clock*(1000/sfrq_mul)/8)
	.endif

sfbl_tbl:
	.dw sfbl_12*2
	.dw sfbl_12*2
	.dw sfbl_8*2
sfbs_tbl:
	.dw sfbs_12*2
	.dw sfbs_12*2
	.dw sfbs_8*2

;72*bitrate/sfrq
frm_sz_tbl:
	.dw int(72*8/11.025)
	.dw int(72*16/11.025)
	.dw int(72*24/11.025)
	.dw int(72*32/11.025)
	.dw int(72*40/11.025)
	.dw int(72*48/11.025)
	.dw int(72*56/11.025)
	.dw int(72*64/11.025)
	.dw 72*8/12
	.dw 72*16/12
	.dw 72*24/12
	.dw 72*32/12
	.dw 72*40/12
	.dw 72*48/12
	.dw 72*56/12
	.dw 72*64/12
	.dw 72*8/8
	.dw 72*16/8
	.dw 72*24/8
	.dw 72*32/8
	.dw 72*40/8
	.dw 72*48/8
	.dw 72*56/8
	.dw 72*64/8
	.endif

sdctconst:
 .dw 42126	;c3
 .dw 64540	;c0
 .dw 22415	;c4
 .dw 56756	;c1
 .dw 56756	;c1
 .dw 61584	;c6
 .dw 11380	;c5
 .dw 50203	;c2

;2 * cos(PI * (2 * i + 1) / (2 * 18))
sdctscale:
 .dw 65287
 .dw 63303
 .dw 59396
 .dw 53684
 .dw 46341
 .dw 37590
 .dw 27697
 .dw 16962
 .dw 5712

;2 * cos(PI * (2 * i + 1) / (4 * 18))
dctscale:
 .dw 65474
 .dw 64975
 .dw 63983
 .dw 62503
 .dw 60547
 .dw 58131
 .dw 55273
 .dw 51993
 .dw 48318
 .dw 44275
 .dw 39896
 .dw 35212
 .dw 30261
 .dw 25080
 .dw 19707
 .dw 14185
 .dw 8554
 .dw 2859

window_l:
 .dw 2859
 .dw 8554
 .dw 14185
 .dw 19707
 .dw 25080
 .dw 30261
 .dw 35212
 .dw 39896
 .dw 44275
 .dw 48318
 .dw 51993
 .dw 55273
 .dw 58131
 .dw 60547
 .dw 62503
 .dw 63983
 .dw 64975
 .dw 65474
 .dw 65474
 .dw 64975
 .dw 63983
 .dw 62503
 .dw 60547
 .dw 58131
 .dw 55273
 .dw 51993
 .dw 48318
 .dw 44275
 .dw 39896
 .dw 35212
 .dw 30261
 .dw 25080
 .dw 19707
 .dw 14185
 .dw 8554
 .dw 2859

window_s:
 .dw 8554
 .dw 25080
 .dw 39896
 .dw 51993
 .dw 60547
 .dw 64975
 .dw 64975
 .dw 60547
 .dw 51993
 .dw 39896
 .dw 25080
 .dw 8554

div5tab:
	.db $00,$01,$02,$03,$04,$10,$11,$12,$13,$14
	.db $20,$21,$22,$23,$24,$30,$31,$32,$33,$34
	.db $40,$41,$42,$43,$44								,-1
	
div3tab:
	.db $00,$01,$02,$10,$11,$12,$20,$21,$22,$30,$31		,-1

nrtab_l:
	.db 6,5,5,5		
	.db 6,5,7,3
	.db 11,10,0,0
nrtab_s:
	.db 9,9,9,9
	.db 9,9,12,6
	.db 18,18,0,0

	.if eval_hd==1
sfbl_8: .dw 6,12,18,24,30,36,44,54,66,80,96,116,140,168,200,238,283,284,285,286,287,288
sfbl_12: .dw 3,6,9,12,15,18,22,27,33,40,48,58,70,84,100,119,142,168,198,232,261,288
sfbs_8: .dw 4,8,12,16,20,24,28,32,36,42,48,54,62,70,78,88,98,108,120,132,144,158,172,186,204,222,240,241,242,243,244,245,246,247,248,249,262,275,288
sfbs_12: .dw 2,4,6,8,10,12,14,16,18,21,24,27,31,35,39,44,49,54,60,66,72,79,86,93,102,111,120,132,144,156,171,186,201,221,241,261,270,279,288
	.else
	.if sfrq==8
sfbl: .dw 6,12,18,24,30,36,44,54,66,80,96,116,140,168,200,238,283,284,285,286,287,288
	.endif
	.if sfrq==12
sfbl: .dw 3,6,9,12,15,18,22,27,33,40,48,58,70,84,100,119,142,168,198,232,261,288
	.endif
	.endif

reqtabscfac:
 .dw 65535 ;2^(0)
 .dw 46341 ;2^(-0.5)
 .dw 32768 ;2^(-1)
 .dw 23170 ;2^(-1.5)
 .dw 16384 ;2^(-2)
 .dw 11585 ;2^(-2.5)
 .dw 8192 ;2^(-3)
 .dw 5793 ;2^(-3.5)
 .dw 4096 ;2^(-4)
 .dw 2896 ;2^(-4.5)
 .dw 2048 ;2^(-5)
 .dw 1448 ;2^(-5.5)
 .dw 1024 ;2^(-6)
 .dw 724 ;2^(-6.5)
 .dw 512 ;2^(-7)
 .dw 362 ;2^(-7.5)
reqtabscfac1:
 .dw 65535 ;2^(0)
 .dw 32768 ;2^(-1)
 .dw 16384 ;2^(-2)
 .dw 8192 ;2^(-3)
 .dw 4096 ;2^(-4)
 .dw 2048 ;2^(-5)
 .dw 1024 ;2^(-6)
 .dw 512 ;2^(-7)
 .dw 256 ;2^(-8)
 .dw 128 ;2^(-9)
 .dw 64 ;2^(-10)
 .dw 32 ;2^(-11)
 .dw 16 ;2^(-12)
 .dw 8 ;2^(-13)
 .dw 4 ;2^(-14)
 .dw 2 ;2^(-15)

reqtabgain:
 .dw int(32768*0.0662902)
 .dw int(32768*0.0788328)
 .dw int(32768*0.0937485)
 .dw int(32768*0.111486)
 .dw int(32768*0.13258)
 .dw int(32768*0.157666)
 .dw int(32768*0.187497)
 .dw int(32768*0.222973)
 .dw int(32768*0.265161)
 .dw int(32768*0.315331)
 .dw int(32768*0.374994)
 .dw int(32768*0.445946)
 .dw int(32768*0.530322)
 .dw int(32768*0.630663)
 .dw int(32768*0.749988)
 .dw int(32768*0.891891)
 .dw int(32768*1.06064)
 .dw int(32768*1.26133)
 .dw int(32768*1.49998)
 .dw int(32768*1.78378)
 .dw 2 ;152
 .dw 3 ;153
 .dw 3 ;154
 .dw 4 ;155
 .dw 4 ;156
 .dw 5 ;157
 .dw 6 ;158
 .dw 7 ;159
 .dw 8 ;160
 .dw 10 ;161
 .dw 12 ;162
 .dw 14 ;163
 .dw 17 ;164
 .dw 20 ;165
 .dw 24 ;166
 .dw 29 ;167
 .dw 34 ;168
 .dw 40 ;169
 .dw 48 ;170
 .dw 57 ;171
 .dw 68 ;172
 .dw 81 ;173
 .dw 96 ;174
 .dw 114 ;175
 .dw 136 ;176
 .dw 161 ;177
 .dw 192 ;178
 .dw 228 ;179
 .dw 272 ;180
 .dw 323 ;181
 .dw 384 ;182
 .dw 457 ;183
 .dw 543 ;184
 .dw 646 ;185
 .dw 768 ;186
 .dw 913 ;187
 .dw 1086 ;188
 .dw 1292 ;189
 .dw 1536 ;190
 .dw 1827 ;191
 .dw 2172 ;192
 .dw 2583 ;193
 .dw 3072 ;194
 .dw 3653 ;195
 .dw 4344 ;196
 .dw 5166 ;197
 .dw 6144 ;198
 .dw 7306 ;199
 .dw 8689 ;200
 .dw 10333 ;201
 .dw 12288 ;202
 .dw 14613 ;203
 .dw 17378 ;204
 .dw 20666 ;205
 .dw 24576 ;206
 .dw 29226 ;207
 .dw 34755 ;208
 .dw 41331 ;209
max_gain: .dw 49151 ;210

csa:
 .dw $6dc2 ;+0.857492926
 .dw $be25 ;-0.514495755

 .dw $70dd ;+0.881741997
 .dw $c39e ;-0.471731969

 .dw $798d ;+0.949628649
 .dw $d7e3 ;-0.313377454

 .dw $7ddd ;+0.983314592
 .dw $e8b7 ;-0.181913200

 .dw $7f6d ;+0.995517816
 .dw $f3e5 ;-0.094574193

 .dw $7fe4 ;+0.999160558
 .dw $fac2 ;-0.040965583

 .dw $7ffc ;+0.999899195
 .dw $fe2f ;-0.014198569

 .dw $7fff ;+0.999993155
 .dw $ff87 ;-0.003699975

imdct_s_tbl:
;0
 .dw int(0.608761429*32768)
 .dw int(-0.923879533*32768)
 .dw int(-0.130526192*32768)
 .dw int(0.991444861*32768)
 .dw int(-0.382683432*32768)
 .dw int(-0.793353340*32768)
;6
 .dw int(-0.793353340*32768)
 .dw int(0.382683432*32768)
 .dw int(0.991444861*32768)
 .dw int(0.130526192*32768)
 .dw int(-0.923879533*32768)
 .dw int(-0.608761429*32768)
;1
 .dw int(0.382683432*32768)
 .dw int(-0.923879533*32768)
 .dw int(0.923879533*32768)
 .dw int(-0.382683432*32768)
 .dw int(-0.382683432*32768)
 .dw int(0.923879533*32768)
;7
 .dw int(-0.923879533*32768)
 .dw int(-0.382683432*32768)
 .dw int(0.382683432*32768)
 .dw int(0.923879533*32768)
 .dw int(0.923879533*32768)
 .dw int(0.382683432*32768)
;2
 .dw int(0.130526192*32768)
 .dw int(-0.382683432*32768)
 .dw int(0.608761429*32768)
 .dw int(-0.793353340*32768)
 .dw int(0.923879533*32768)
 .dw int(-0.991444861*32768)
;8
 .dw int(-0.991444861*32768)
 .dw int(-0.923879533*32768)
 .dw int(-0.793353340*32768)
 .dw int(-0.608761429*32768)
 .dw int(-0.382683432*32768)
 .dw int(-0.130526192*32768)

	.include "powertab.asm"

htables:
	.dw -1					;not used
	.dw tbl1-PC-1			;table 1, linbits=0
	.dw tbl2-PC-1			;table 2, linbits=0
	.dw tbl3-PC-1			;table 3, linbits=0
	.dw -1					;not used
	.dw tbl5-PC-1			;table 5, linbits=0
	.dw tbl6-PC-1			;table 6, linbits=0
	.dw tbl7-PC-1			;table 7, linbits=0
	.dw tbl8-PC-1			;table 8, linbits=0
	.dw tbl9-PC-1			;table 9, linbits=0
	.dw tbl10-PC-1			;table 10, linbits=0
	.dw tbl11-PC-1			;table 11, linbits=0
	.dw tbl12-PC-1			;table 12, linbits=0
	.dw tbl13-PC-1			;table 13, linbits=0
	.dw -1					;not used
	.dw tbl15-PC-1			;table 15, linbits=0
	.dw 1<<12|(tbl16-PC-1)	;table 16, linbits=1
	.dw 2<<12|(tbl16-PC-1)	;table 17, linbits=2
	.dw 3<<12|(tbl16-PC-1)	;table 18, linbits=3
	.dw 4<<12|(tbl16-PC-1)	;table 19, linbits=4
	.dw 6<<12|(tbl16-PC-1)	;table 20, linbits=6
	.dw 8<<12|(tbl16-PC-1)	;table 21, linbits=8
	.dw 10<<12|(tbl16-PC-1)	;table 22, linbits=10
	.dw 13<<12|(tbl16-PC-1)	;table 23, linbits=13
	.dw 4<<12|(tbl24-PC-1)	;table 24, linbits=4
	.dw 5<<12|(tbl24-PC-1)	;table 25, linbits=5
	.dw 6<<12|(tbl24-PC-1)	;table 26, linbits=6
	.dw 7<<12|(tbl24-PC-1)	;table 27, linbits=7
	.dw 8<<12|(tbl24-PC-1)	;table 28, linbits=8
	.dw 9<<12|(tbl24-PC-1)	;table 29, linbits=9
	.dw 11<<12|(tbl24-PC-1)	;table 30, linbits=11
	.dw 13<<12|(tbl24-PC-1)	;table 31, linbits=13

	.include "Table1.asm"
	.include "Table2.asm"
	.include "Table3.asm"
	.include "Table5.asm"
	.include "Table6.asm"
	.include "Table7.asm"
	.include "Table8.asm"
	.include "Table9.asm"
	.include "Table10.asm"
	.include "Table11.asm"
	.include "Table12.asm"
	.include "Table13.asm"
	.include "Table15.asm"
	.include "Table16.asm"
	.include "Table24.asm"
	.include "TableA.asm"

synthcoeff:
; 0
 .dw 1
 .dw -7
 .dw 53
 .dw -115
 .dw 509
 .dw -1288
 .dw 1643
 .dw -9372
 .dw 18760
 .dw 9372
 .dw 1643
 .dw 1288
 .dw 509
 .dw 115
 .dw 53
 .dw 7
 .dw 1
 .dw -7
 .dw 53
 .dw -115
 .dw 509
 .dw -1288
 .dw 1643
 .dw -9372
 .dw 18760
 .dw 9372
 .dw 1643
 .dw 1288
 .dw 509
 .dw 115
 .dw 53
 .dw 7
; 1
 .dw -1
 .dw -8
 .dw 54
 .dw -130
 .dw 500
 .dw -1379
 .dw 1490
 .dw -9834
 .dw 18748
 .dw 8910
 .dw 1783
 .dw 1197
 .dw 516
 .dw 100
 .dw 52
 .dw 7
 .dw -1
 .dw -8
 .dw 54
 .dw -130
 .dw 500
 .dw -1379
 .dw 1490
 .dw -9834
 .dw 18748
 .dw 8910
 .dw 1783
 .dw 1197
 .dw 516
 .dw 100
 .dw 52
 .dw 7
; 2
 .dw -1
 .dw -9
 .dw 55
 .dw -145
 .dw 488
 .dw -1470
 .dw 1322
 .dw -10294
 .dw 18714
 .dw 8448
 .dw 1910
 .dw 1106
 .dw 520
 .dw 87
 .dw 50
 .dw 6
 .dw -1
 .dw -9
 .dw 55
 .dw -145
 .dw 488
 .dw -1470
 .dw 1322
 .dw -10294
 .dw 18714
 .dw 8448
 .dw 1910
 .dw 1106
 .dw 520
 .dw 87
 .dw 50
 .dw 6
; 3
 .dw -1
 .dw -10
 .dw 56
 .dw -161
 .dw 473
 .dw -1559
 .dw 1140
 .dw -10751
 .dw 18658
 .dw 7987
 .dw 2023
 .dw 1016
 .dw 522
 .dw 74
 .dw 49
 .dw 5
 .dw -1
 .dw -10
 .dw 56
 .dw -161
 .dw 473
 .dw -1559
 .dw 1140
 .dw -10751
 .dw 18658
 .dw 7987
 .dw 2023
 .dw 1016
 .dw 522
 .dw 74
 .dw 49
 .dw 5
; 4
 .dw -1
 .dw -10
 .dw 57
 .dw -178
 .dw 456
 .dw -1647
 .dw 944
 .dw -11205
 .dw 18578
 .dw 7528
 .dw 2123
 .dw 926
 .dw 521
 .dw 61
 .dw 48
 .dw 5
 .dw -1
 .dw -10
 .dw 57
 .dw -178
 .dw 456
 .dw -1647
 .dw 944
 .dw -11205
 .dw 18578
 .dw 7528
 .dw 2123
 .dw 926
 .dw 521
 .dw 61
 .dw 48
 .dw 5
; 5
 .dw -1
 .dw -11
 .dw 57
 .dw -195
 .dw 435
 .dw -1734
 .dw 734
 .dw -11654
 .dw 18477
 .dw 7072
 .dw 2210
 .dw 838
 .dw 519
 .dw 49
 .dw 46
 .dw 4
 .dw -1
 .dw -11
 .dw 57
 .dw -195
 .dw 435
 .dw -1734
 .dw 734
 .dw -11654
 .dw 18477
 .dw 7072
 .dw 2210
 .dw 838
 .dw 519
 .dw 49
 .dw 46
 .dw 4
; 6
 .dw -1
 .dw -12
 .dw 57
 .dw -212
 .dw 411
 .dw -1818
 .dw 509
 .dw -12098
 .dw 18354
 .dw 6621
 .dw 2285
 .dw 751
 .dw 514
 .dw 38
 .dw 44
 .dw 4
 .dw -1
 .dw -12
 .dw 57
 .dw -212
 .dw 411
 .dw -1818
 .dw 509
 .dw -12098
 .dw 18354
 .dw 6621
 .dw 2285
 .dw 751
 .dw 514
 .dw 38
 .dw 44
 .dw 4
; 7
 .dw -1
 .dw -13
 .dw 57
 .dw -230
 .dw 384
 .dw -1899
 .dw 271
 .dw -12534
 .dw 18209
 .dw 6173
 .dw 2347
 .dw 666
 .dw 508
 .dw 28
 .dw 42
 .dw 3
 .dw -1
 .dw -13
 .dw 57
 .dw -230
 .dw 384
 .dw -1899
 .dw 271
 .dw -12534
 .dw 18209
 .dw 6173
 .dw 2347
 .dw 666
 .dw 508
 .dw 28
 .dw 42
 .dw 3
; 8
 .dw -1
 .dw -15
 .dw 56
 .dw -248
 .dw 354
 .dw -1977
 .dw 17
 .dw -12963
 .dw 18042
 .dw 5732
 .dw 2398
 .dw 583
 .dw 500
 .dw 18
 .dw 40
 .dw 3
 .dw -1
 .dw -15
 .dw 56
 .dw -248
 .dw 354
 .dw -1977
 .dw 17
 .dw -12963
 .dw 18042
 .dw 5732
 .dw 2398
 .dw 583
 .dw 500
 .dw 18
 .dw 40
 .dw 3
; 9
 .dw -1
 .dw -16
 .dw 55
 .dw -266
 .dw 320
 .dw -2052
 .dw -249
 .dw -13384
 .dw 17855
 .dw 5297
 .dw 2437
 .dw 502
 .dw 490
 .dw 9
 .dw 39
 .dw 3
 .dw -1
 .dw -16
 .dw 55
 .dw -266
 .dw 320
 .dw -2052
 .dw -249
 .dw -13384
 .dw 17855
 .dw 5297
 .dw 2437
 .dw 502
 .dw 490
 .dw 9
 .dw 39
 .dw 3
; 10
 .dw -1
 .dw -17
 .dw 54
 .dw -284
 .dw 283
 .dw -2123
 .dw -530
 .dw -13795
 .dw 17648
 .dw 4869
 .dw 2466
 .dw 423
 .dw 480
 .dw 1
 .dw 37
 .dw 3
 .dw -1
 .dw -17
 .dw 54
 .dw -284
 .dw 283
 .dw -2123
 .dw -530
 .dw -13795
 .dw 17648
 .dw 4869
 .dw 2466
 .dw 423
 .dw 480
 .dw 1
 .dw 37
 .dw 3
; 11
 .dw -1
 .dw -18
 .dw 52
 .dw -303
 .dw 242
 .dw -2189
 .dw -825
 .dw -14194
 .dw 17420
 .dw 4450
 .dw 2484
 .dw 347
 .dw 468
 .dw -7
 .dw 35
 .dw 2
 .dw -1
 .dw -18
 .dw 52
 .dw -303
 .dw 242
 .dw -2189
 .dw -825
 .dw -14194
 .dw 17420
 .dw 4450
 .dw 2484
 .dw 347
 .dw 468
 .dw -7
 .dw 35
 .dw 2
; 12
 .dw -1
 .dw -20
 .dw 50
 .dw -321
 .dw 199
 .dw -2250
 .dw -1133
 .dw -14583
 .dw 17173
 .dw 4039
 .dw 2492
 .dw 274
 .dw 454
 .dw -14
 .dw 33
 .dw 2
 .dw -1
 .dw -20
 .dw 50
 .dw -321
 .dw 199
 .dw -2250
 .dw -1133
 .dw -14583
 .dw 17173
 .dw 4039
 .dw 2492
 .dw 274
 .dw 454
 .dw -14
 .dw 33
 .dw 2
; 13
 .dw -1
 .dw -21
 .dw 47
 .dw -339
 .dw 151
 .dw -2305
 .dw -1455
 .dw -14960
 .dw 16907
 .dw 3637
 .dw 2490
 .dw 203
 .dw 440
 .dw -21
 .dw 31
 .dw 2
 .dw -1
 .dw -21
 .dw 47
 .dw -339
 .dw 151
 .dw -2305
 .dw -1455
 .dw -14960
 .dw 16907
 .dw 3637
 .dw 2490
 .dw 203
 .dw 440
 .dw -21
 .dw 31
 .dw 2
; 14
 .dw -1
 .dw -23
 .dw 44
 .dw -357
 .dw 100
 .dw -2354
 .dw -1789
 .dw -15322
 .dw 16624
 .dw 3245
 .dw 2479
 .dw 136
 .dw 425
 .dw -27
 .dw 29
 .dw 2
 .dw -1
 .dw -23
 .dw 44
 .dw -357
 .dw 100
 .dw -2354
 .dw -1789
 .dw -15322
 .dw 16624
 .dw 3245
 .dw 2479
 .dw 136
 .dw 425
 .dw -27
 .dw 29
 .dw 2
; 15
 .dw -1
 .dw -24
 .dw 41
 .dw -374
 .dw 46
 .dw -2396
 .dw -2135
 .dw -15671
 .dw 16323
 .dw 2864
 .dw 2460
 .dw 72
 .dw 408
 .dw -32
 .dw 28
 .dw 2
 .dw -1
 .dw -24
 .dw 41
 .dw -374
 .dw 46
 .dw -2396
 .dw -2135
 .dw -15671
 .dw 16323
 .dw 2864
 .dw 2460
 .dw 72
 .dw 408
 .dw -32
 .dw 28
 .dw 2
; 16
 .dw -1
 .dw -26
 .dw 36
 .dw -392
 .dw -11
 .dw -2432
 .dw -2494
 .dw -16005
 .dw 16005
 .dw 2494
 .dw 2432
 .dw 11
 .dw 392
 .dw -36
 .dw 26
 .dw 1
 .dw -1
 .dw -26
 .dw 36
 .dw -392
 .dw -11
 .dw -2432
 .dw -2494
 .dw -16005
 .dw 16005
 .dw 2494
 .dw 2432
 .dw 11
 .dw 392
 .dw -36
 .dw 26
 .dw 1

	.if flash!=1
mp3file:
;	.include "FA_noshort_nores_8khz_16kbits.inc"
;	.include "FA_noshort_nores_12khz_40kbits.inc"
;	.include "FA_nores_8khz_16kbits.inc"
;	.include "FA_nores_11.025khz_40kbits.inc"
;	.include "FA_nores_12khz_40kbits.inc"
;	.include "FA_8khz_16kbits.inc"
;	.include "FA_11.025khz_40kbits.inc"
;	.include "FA_12khz_40kbits.inc"
	.include "FA_8khz_vbr.inc"
;	.include "FA_11.025khz_vbr.inc"
;	.include "FA_12khz_vbr.inc"
mp3file_end:
	.endif

	.dseg
	.org SRAM_START
#ifdef atxmega
	.if br_support
;frame_buf MUST start at a boundary aligned with fr_buf_size
frame_buf: .byte fr_buf_size
	.endif
lo:		.byte 2*16*16
hi:		.byte 2*16*16
frame:	.byte 2*576
ovlb: 	.byte 2*576
sample: .byte 2*576
	.if (chip==1 || flash==1) && use_dac!=1 && eval_hd!=1
output:	.byte 576
	.else
output: .byte 2*576
	.endif
output1: .byte 2*576
	.if eval_hd
frame_sz: .byte 2
sfb_ptr: .byte 2
	.endif
	.if br_support
fr_wr_ptr: .byte 2
fr_rd_ptr: .byte 2
	.endif
phase:	.byte 1
outptr:	.byte 2
part23len: .byte 2
bigv:	.byte 2
ggain:	.byte 1
scfacc:	.byte 2
table1:	.byte 1
table2:	.byte 1
table3:	.byte 1
r0cnt:	.byte 1
r1cnt:	.byte 1
scfacsc: .byte 1
cnt1tbl: .byte 1
	.if eval_hd==1
sblk_gain: .byte 3
	.endif

a:	.byte 26*2
m:	.byte 8*2
tmp: .byte 9*2
tmp1: .byte 18*2
tmp2: .byte 18*2
_z:	.byte 36*2

.equ slen=a				;.byte 4

.equ scfac=a+4			;.byte 39

.equ scfac1=scfac+39	;.byte 39*3

.equ tm32=part23len		;.byte 2*32
.equ tm64=tm32+64		;.byte 2*32

#else
;
;use for ATmega1284
;
sample: .byte 2*576
frame:	.byte 2*576
ovlb: 	.byte 2*576
	.if eval_hd
frame_sz: .byte 2
sfb_ptr: .byte 2
	.endif
	.if br_support
fr_wr_ptr: .byte 2
fr_rd_ptr: .byte 2
	.endif
phase:	.byte 1
outptr:	.byte 2
part23len: .byte 2
bigv:	.byte 2
ggain:	.byte 1
scfacc:	.byte 2
table1:	.byte 1
table2:	.byte 1
table3:	.byte 1
r0cnt:	.byte 1
r1cnt:	.byte 1
scfacsc: .byte 1
cnt1tbl: .byte 1
	.if eval_hd==1
sblk_gain: .byte 3
	.endif

a:	.byte 26*2
m:	.byte 8*2
tmp: .byte 9*2
tmp1: .byte 18*2
tmp2: .byte 18*2
_z:	.byte 36*2

.equ slen=a				;.byte 4

.equ scfac=a+4			;.byte 39

.equ scfac1=scfac+39	;.byte 39*3

.equ tm32=part23len		;.byte 2*32
.equ tm64=tm32+64		;.byte 2*32
tptr:
;stack will be placed below frame_buf if br_support enabled
	.org (tptr+fr_buf_size-1) & ~(fr_buf_size-1)
	.if br_support
;frame_buf MUST start at a boundary aligned with fr_buf_size
frame_buf: .byte fr_buf_size
	.endif
lo:		.byte 2*16*16
hi:		.byte 2*16*16

	.if (chip==1 || flash==1) && use_dac!=1 && eval_hd!=1
output:	.byte 576
	.else
output: .byte 2*576
	.endif
output1: .byte 2*576


/*
;
;use for ATmega 640/1280/1281/2560/2561
;if used delete the definitions for ATmega1284 above
;
sample: .byte 2*576
frame:	.byte 2*576
ovlb: 	.byte 2*576
_z:	.byte 36*2
tptr:
;stack will be placed below frame_buf if br_support enabled
	.org (tptr+fr_buf_size-1) & ~(fr_buf_size-1)
	.if br_support
;frame_buf MUST start at a boundary aligned with fr_buf_size
frame_buf: .byte fr_buf_size
	.endif
lo:		.byte 2*16*16
hi:		.byte 2*16*16

	.if (chip==1 || flash==1) && use_dac!=1 && eval_hd!=1
output:	.byte 576
	.else
output: .byte 2*576
	.endif
output1: .byte 2*576
	.if eval_hd
frame_sz: .byte 2
sfb_ptr: .byte 2
	.endif
	.if br_support
fr_wr_ptr: .byte 2
fr_rd_ptr: .byte 2
	.endif
phase:	.byte 1
outptr:	.byte 2
part23len: .byte 2
bigv:	.byte 2
ggain:	.byte 1
scfacc:	.byte 2
table1:	.byte 1
table2:	.byte 1
table3:	.byte 1
r0cnt:	.byte 1
r1cnt:	.byte 1
scfacsc: .byte 1
cnt1tbl: .byte 1
	.if eval_hd==1
sblk_gain: .byte 3
	.endif

a:	.byte 26*2
m:	.byte 8*2
tmp: .byte 9*2
tmp1: .byte 18*2
tmp2: .byte 18*2

.equ slen=a				;.byte 4

.equ scfac=a+4			;.byte 39

.equ scfac1=scfac+39	;.byte 39*3

.equ tm32=part23len		;.byte 2*32
.equ tm64=tm32+64		;.byte 2*32
*/
#endif
